# RCAIDE/Library/Missions/Segments/Common/Update/dimensionless.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Update Differentials Altitude
# ---------------------------------------------------------------------------------------------------------------------- 
def differentials_altitude(segment):
    """ Initializes the differntial altitude
    
        Assumptions: 
            None
        
        Inputs: 
            state.conditions:           
                numerics.dimensionless.integrate        [-]
                numerics.dimensionless.control_points   [s]
                frames.position_vector                  [m]
                inertial.velocity_vector                [m/s]
            
        Outputs:
            state.conditions.frames.inertial.time       [s]
           
        Properties Used:
        N/A 
    """

    # unpack
    t = segment.state.numerics.dimensionless.control_points 
    I = segment.state.numerics.dimensionless.integrate
    r = segment.state.conditions.frames.inertial.position_vector
    v = segment.state.conditions.frames.inertial.velocity_vector

    dz = r[-1,2] - r[0,2]
    vz = v[:,2,None] # maintain column array

    # get overall time step
    dt = np.dot( I[-1,:] * dz , 1/ vz[:,0] )

    # rescale operators
    t = t * dt

    # pack
    t_initial = segment.state.conditions.frames.inertial.time[0,0]
    segment.state.conditions.frames.inertial.time[:,0] = t_initial + t[:,0]

    return


# RCAIDE/Framework/External_Interfaces/OpenVSP/vsp_boom.py

# Created:  Jun 2018, T. St Francis
# Modified: Aug 2018, T. St Francis
#           Jan 2020, T. MacDonald
#           Jul 2020, E. Botero

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports 
import RCAIDE
from RCAIDE.Framework.Core import Units, Data  
import numpy as np
try:
    import vsp as vsp
except ImportError:
    try:
        import openvsp as vsp
    except ImportError:
        # This allows RCAIDE to build without OpenVSP
        pass
    
# ---------------------------------------------------------------------------------------------------------------------- 
#  vsp read boom
# ---------------------------------------------------------------------------------------------------------------------- 
def read_vsp_boom(b_id,fux_idx,sym_flag, units_type='SI', fineness=True, use_scaling=True):
    """This reads an OpenVSP boom geometry and writes it to a RCAIDE boom format.

    Assumptions:
    1. OpenVSP boom is "conventionally shaped" (generally narrow at nose and tail, wider in center). 
    2. Boom is designed in VSP as it appears in real life. That is, the VSP model does not rely on
       superficial elements such as canopies, stacks, or additional booms to cover up internal lofting oddities.
    3. This program will NOT account for multiple geometries comprising the boom. For example: a wingbox mounted beneath
       is a separate geometry and will NOT be processed.
    4. Boom origin is located at nose. VSP file origin can be located anywhere, preferably at the forward tip
       of the vehicle or in front (to make all X-coordinates of vehicle positive).
    5. Written for OpenVSP 3.21.1

    Source:
    N/A

    Inputs:
    0. Pre-loaded VSP vehicle in memory, via import_vsp_vehicle.
    1. VSP 10-digit geom ID for boom.
    2. Units_type set to 'SI' (default) or 'Imperial'.
    3. Boolean for whether or not to compute boom finenesses (default = True).
    4. Boolean for whether or not to use the scaling from OpenVSP (default = True).

    Outputs:

    Writes RCAIDE boom, with these geometries:           (all defaults are SI, but user may specify Imperial)

    	Booms.Boom.			
    		origin                                  [m] in all three dimensions
    		width                                   [m]
    		lengths.
    		  total                                 [m]
    		  nose                                  [m]
    		  tail                                  [m]
    		heights.
    		  maximum                               [m]
    		  at_quarter_length                     [m]
    		  at_three_quarters_length              [m]
    		effective_diameter                      [m]
    		fineness.nose                           [-] ratio of nose section length to boom effective diameter
    		fineness.tail                           [-] ratio of tail section length to boom effective diameter
    		areas.wetted                            [m^2]
    		tag                                     <string>
    		segment[].   (segments are in ordered container and callable by number)
    		  vsp.shape                               [point,circle,round_rect,general_fuse,fuse_file]
    		  vsp.xsec_id                             <10 digit string>
    		  percent_x_location
    		  percent_z_location
    		  height
    		  width
    		  length
    		  effective_diameter
    		  tag
    		vsp.xsec_num                              <integer of boom segment quantity>
    		vsp.xsec_surf_id                          <10 digit string>

    Properties Used:
    N/A
    """  	
    boom = RCAIDE.Library.Components.Booms.Boom()	

    if units_type == 'SI':
        units_factor = Units.meter * 1.
    elif units_type == 'imperial':
        units_factor = Units.foot * 1.
    elif units_type == 'inches':
        units_factor = Units.inch * 1.	 
     
    if vsp.GetGeomName(b_id):
        boom.tag = vsp.GetGeomName(b_id) + '_' + str(fux_idx+1)
    else: 
        boom.tag = 'BoomGeom' + '_' + str(fux_idx+1)	
        
    if use_scaling:
        scaling       = vsp.GetParmVal(b_id, 'Scale', 'XForm')  
    else:
        scaling       = 1.
    units_factor      = units_factor*scaling

    boom.origin[0][0] = vsp.GetParmVal(b_id, 'X_Location', 'XForm') * units_factor
    boom.origin[0][1] = vsp.GetParmVal(b_id, 'Y_Location', 'XForm') * units_factor*sym_flag
    boom.origin[0][2] = vsp.GetParmVal(b_id, 'Z_Location', 'XForm') * units_factor

    boom.lengths.total         = vsp.GetParmVal(b_id, 'Length', 'Design') * units_factor
    boom.vsp_data.xsec_surf_id = vsp.GetXSecSurf(b_id, 0) 			        # There is only one XSecSurf in geom.
    boom.vsp_data.xsec_num     = vsp.GetNumXSec(boom.vsp_data.xsec_surf_id) 		# Number of xsecs in boom.	 

        
    x_locs    = []
    heights   = []
    widths    = []
    eff_diams = []
    lengths   = []

    # -----------------
    # Boom segments
    # -----------------

    for ii in range(0, boom.vsp_data.xsec_num): 
        # Create the segment
        x_sec                     = vsp.GetXSec(boom.vsp_data.xsec_surf_id, ii) # VSP XSec ID.
        segment                   = RCAIDE.Library.Components.Booms.Segment()
        segment.vsp_data.xsec_id  = x_sec 
        segment.tag               = 'segment_' + str(ii)

        # Pull out Parms that will be needed
        X_Loc_P = vsp.GetXSecParm(x_sec, 'XLocPercent')
        Z_Loc_P = vsp.GetXSecParm(x_sec, 'ZLocPercent')

        segment.percent_x_location = vsp.GetParmVal(X_Loc_P) # Along boom length.
        segment.percent_z_location = vsp.GetParmVal(Z_Loc_P ) # Vertical deviation of boom center.
        segment.height             = vsp.GetXSecHeight(segment.vsp_data.xsec_id) * units_factor
        segment.width              = vsp.GetXSecWidth(segment.vsp_data.xsec_id) * units_factor
        segment.effective_diameter = (segment.height+segment.width)/2. 

        x_locs.append(segment.percent_x_location)	 # Save into arrays for later computation.
        heights.append(segment.height)
        widths.append(segment.width)
        eff_diams.append(segment.effective_diameter)

        if ii != (boom.vsp_data.xsec_num-1): # Segment length: stored as length since previous segment. (last segment will have length 0.0.)
            next_xsec = vsp.GetXSec(boom.vsp_data.xsec_surf_id, ii+1)
            X_Loc_P_p = vsp.GetXSecParm(next_xsec, 'XLocPercent')
            percent_x_loc_p1 = vsp.GetParmVal(X_Loc_P_p) 
            segment.length = boom.lengths.total*(percent_x_loc_p1 - segment.percent_x_location) * units_factor
        else:
            segment.length = 0.0
        lengths.append(segment.length)

        shape	   = vsp.GetXSecShape(segment.vsp_data.xsec_id)
        shape_dict = {0:'point',1:'circle',2:'ellipse',3:'super ellipse',4:'rounded rectangle',5:'general fuse',6:'fuse file'}
        segment.vsp_data.shape = shape_dict[shape]	

        boom.segments.append(segment)

    boom.heights.at_quarter_length          = get_boom_height(boom, .25)  # Calls get_boom_height function (below).
    boom.heights.at_three_quarters_length   = get_boom_height(boom, .75) 
    boom.heights.at_wing_root_quarter_chord = get_boom_height(boom, .4) 

    boom.heights.maximum    = max(heights)          # Max segment height.	
    boom.width              = max(widths)           # Max segment width.
    boom.effective_diameter = max(eff_diams)        # Max segment effective diam.

    boom.areas.front_projected  = np.pi*((boom.effective_diameter)/2)**2

    eff_diam_gradients_fwd = np.array(eff_diams[1:]) - np.array(eff_diams[:-1])		# Compute gradients of segment effective diameters.
    eff_diam_gradients_fwd = np.multiply(eff_diam_gradients_fwd, lengths[:-1])

    boom = compute_boom_fineness(boom, x_locs, eff_diams, eff_diam_gradients_fwd)	

    return boom

# ---------------------------------------------------------------------------------------------------------------------- 
# Write VSP boom
# ---------------------------------------------------------------------------------------------------------------------- 
def write_vsp_boom(boom,area_tags, OML_set_ind):
    """This writes a boom into OpenVSP format.

    Assumptions:
    None

    Source:
    N/A

    Inputs:
    boom
      width                                   [m]
      lengths.total                           [m]
      heights.
        maximum                               [m]
        at_quarter_length                     [m]
        at_wing_root_quarter_chord            [m]
        at_three_quarters_length              [m]
      effective_diameter                      [m]
      fineness.nose                           [-] ratio of nose section length to boom width
      fineness.tail                           [-] ratio of tail section length to boom width
      tag                                     <string>
      OpenVSP_values.  (optional)
        nose.top.angle                        [degrees]
        nose.top.strength                     [-] this determines how much the specified angle influences that shape
        nose.side.angle                       [degrees]
        nose.side.strength                    [-]
        nose.TB_Sym                           <boolean> determines if top angle is mirrored on bottom
        nose.z_pos                            [-] z position of the nose as a percentage of boom length (.1 is 10%)
        tail.top.angle                        [degrees]
        tail.top.strength                     [-]
        tail.z_pos (optional, 0.02 default)   [-] z position of the tail as a percentage of boom length (.1 is 10%)
      Segments. (optional)
        width                                 [m]
        height                                [m]
        percent_x_location                    [-] .1 is 10% length
        percent_z_location                    [-] .1 is 10% length
    area_tags                                 <dict> used to keep track of all tags needed in wetted area computation           
    main_wing.origin                          [m]
    main_wing.chords.root                     [m]
    fuel_tank_set_index                       <int> OpenVSP object set containing the fuel tanks    

    Outputs:
    Operates on the active OpenVSP model, no direct output

    Properties Used:
    N/A
    """     

    num_segs           = len(boom.segments)
    length             = boom.lengths.total
    fuse_x             = boom.origin[0][0]    
    fuse_y             = boom.origin[0][1]
    fuse_z             = boom.origin[0][2]
    fuse_x_rotation    = boom.x_rotation   
    fuse_y_rotation    = boom.y_rotation
    fuse_z_rotation    = boom.z_rotation
    
    widths  = []
    heights = []
    x_poses = []
    z_poses = []
    segs = boom.segments
    for seg in segs:
        widths.append(seg.width)
        heights.append(seg.height)
        x_poses.append(seg.percent_x_location)
        z_poses.append(seg.percent_z_location)

    end_ind = num_segs-1

    b_id = vsp.AddGeom("BOOM") 
    vsp.SetGeomName(b_id, boom.tag)
    area_tags[boom.tag] = ['booms',boom.tag]

    tail_z_pos = 0.02 # default value

    # set boom relative location and rotation
    vsp.SetParmVal( b_id,'X_Rel_Rotation','XForm',fuse_x_rotation)
    vsp.SetParmVal( b_id,'Y_Rel_Rotation','XForm',fuse_y_rotation)
    vsp.SetParmVal( b_id,'Z_Rel_Rotation','XForm',fuse_z_rotation)

    vsp.SetParmVal( b_id,'X_Rel_Location','XForm',fuse_x)
    vsp.SetParmVal( b_id,'Y_Rel_Location','XForm',fuse_y)
    vsp.SetParmVal( b_id,'Z_Rel_Location','XForm',fuse_z)


    if 'OpenVSP_values' in boom:        
        vals = boom.OpenVSP_values

        # for wave drag testing
        boom.OpenVSP_ID = b_id

        # Nose
        vsp.SetParmVal(b_id,"TopLAngle","XSec_0",vals.nose.top.angle)
        vsp.SetParmVal(b_id,"TopLStrength","XSec_0",vals.nose.top.strength)
        vsp.SetParmVal(b_id,"RightLAngle","XSec_0",vals.nose.side.angle)
        vsp.SetParmVal(b_id,"RightLStrength","XSec_0",vals.nose.side.strength)
        vsp.SetParmVal(b_id,"TBSym","XSec_0",vals.nose.TB_Sym)
        vsp.SetParmVal(b_id,"ZLocPercent","XSec_0",vals.nose.z_pos)
        if not vals.nose.TB_Sym:
            vsp.SetParmVal(b_id,"BottomLAngle","XSec_0",vals.nose.bottom.angle)
            vsp.SetParmVal(b_id,"BottomLStrength","XSec_0",vals.nose.bottom.strength)           
 
        if 'z_pos' in vals.tail:
            tail_z_pos = vals.tail.z_pos
        else:
            pass # use above default

 
    # OpenVSP vals do not exist:
    vals                   = Data()
    vals.nose              = Data()
    vals.tail              = Data()
    vals.tail.top          = Data()

    vals.nose.z_pos        = 0.0
    vals.tail.top.angle    = 0.0
    vals.tail.top.strength = 0.0

    #if len(np.unique(x_poses)) != len(x_poses):
        #raise ValueError('Duplicate boom section positions detected.')
    vsp.SetParmVal(b_id,"Length","Design",length)
    if num_segs != 5: # reduce to only nose and tail
        vsp.CutXSec(b_id,1) # remove extra default section
        vsp.CutXSec(b_id,1) # remove extra default section
        vsp.CutXSec(b_id,1) # remove extra default section
        for i in range(num_segs-2): # add back the required number of sections
            vsp.InsertXSec(b_id, 0, vsp.XS_ELLIPSE)           
            vsp.Update()
    for i in range(num_segs-2):
        # Bunch sections to allow proper length settings in the next step
        # This is necessary because OpenVSP will not move a section past an adjacent section
        vsp.SetParmVal(b_id, "XLocPercent", "XSec_"+str(i+1),1e-6*(i+1))
        vsp.Update()
    if x_poses[1] < (num_segs-2)*1e-6:
        print('Warning: Second boom section is too close to the nose. OpenVSP model may not be accurate.')
    for i in reversed(range(num_segs-2)):
        # order is reversed because sections are initially bunched in the front and cannot be extended passed the next
        vsp.SetParmVal(b_id, "XLocPercent", "XSec_"+str(i+1),x_poses[i+1])
        vsp.SetParmVal(b_id, "ZLocPercent", "XSec_"+str(i+1),z_poses[i+1])
        vsp.SetParmVal(b_id, "Ellipse_Width", "XSecCurve_"+str(i+1), widths[i+1])
        vsp.SetParmVal(b_id, "Ellipse_Height", "XSecCurve_"+str(i+1), heights[i+1])   
        vsp.Update()             
        set_section_angles(i, vals.nose.z_pos, tail_z_pos, x_poses, z_poses, heights, widths,length,end_ind,b_id)            

    vsp.SetParmVal(b_id, "XLocPercent", "XSec_"+str(0),x_poses[0])
    vsp.SetParmVal(b_id, "ZLocPercent", "XSec_"+str(0),z_poses[0])
    vsp.SetParmVal(b_id, "XLocPercent", "XSec_"+str(end_ind),x_poses[-1])
    vsp.SetParmVal(b_id, "ZLocPercent", "XSec_"+str(end_ind),z_poses[-1])    

    # Tail
    if heights[-1] > 0.: 
        pos = len(heights)-1
        vsp.InsertXSec(b_id, pos-1, vsp.XS_ELLIPSE)
        vsp.Update()
        vsp.SetParmVal(b_id, "Ellipse_Width", "XSecCurve_"+str(pos), widths[-1])
        vsp.SetParmVal(b_id, "Ellipse_Height", "XSecCurve_"+str(pos), heights[-1])
        vsp.SetParmVal(b_id, "XLocPercent", "XSec_"+str(pos),x_poses[-1])
        vsp.SetParmVal(b_id, "ZLocPercent", "XSec_"+str(pos),z_poses[-1])              

        xsecsurf = vsp.GetXSecSurf(b_id,0)
        vsp.ChangeXSecShape(xsecsurf,pos+1,vsp.XS_POINT)
        vsp.Update()           
        vsp.SetParmVal(b_id, "XLocPercent", "XSec_"+str(pos+1),x_poses[-1])
        vsp.SetParmVal(b_id, "ZLocPercent", "XSec_"+str(pos+1),z_poses[-1])     

        # update strengths to make end flat
        vsp.SetParmVal(b_id,"TopRStrength","XSec_"+str(pos), 0.)
        vsp.SetParmVal(b_id,"RightRStrength","XSec_"+str(pos), 0.)
        vsp.SetParmVal(b_id,"BottomRStrength","XSec_"+str(pos), 0.)
        vsp.SetParmVal(b_id,"TopLStrength","XSec_"+str(pos+1), 0.)
        vsp.SetParmVal(b_id,"RightLStrength","XSec_"+str(pos+1), 0.)            

    else:
        vsp.SetParmVal(b_id,"TopLAngle","XSec_"+str(end_ind),vals.tail.top.angle)
        vsp.SetParmVal(b_id,"TopLStrength","XSec_"+str(end_ind),vals.tail.top.strength)
        vsp.SetParmVal(b_id,"AllSym","XSec_"+str(end_ind),1)
        vsp.Update()


    if 'z_pos' in vals.tail:
        tail_z_pos = vals.tail.z_pos
    else:
        pass # use above default
    
    vsp.SetSetFlag(b_id, OML_set_ind, True)

    return area_tags

# ---------------------------------------------------------------------------------------------------------------------- 
# set_section_angles 
# ---------------------------------------------------------------------------------------------------------------------- 
def set_section_angles(i,nose_z,tail_z,x_poses,z_poses,heights,widths,length,end_ind,b_id):
    """Set boom section angles to create a smooth (in the non-technical sense) boom shape.
    Note that i of 0 corresponds to the first section that is not the end point.

    Assumptions:
    May fail to give reasonable angles for very irregularly shaped booms
    Does not work on the nose and tail sections.

    Source:
    N/A

    Inputs:  
    nose_z   [-] # 0.1 is 10% of the boom length
    widths   np.array of [m]
    heights  np.array of [m]
    tail_z   [-] # 0.1 is 10% of the boom length

    Outputs:
    Operates on the active OpenVSP model, no direct output

    Properties Used:
    N/A
    """    
    w0 = widths[i]
    h0 = heights[i]
    x0 = x_poses[i]
    z0 = z_poses[i]   
    w2 = widths[i+2]
    h2 = heights[i+2]
    x2 = x_poses[i+2]
    z2 = z_poses[i+2]

    x0 = x0*length
    x2 = x2*length
    z0 = z0*length
    z2 = z2*length

    top_z_diff = (h2/2+z2)-(h0/2+z0)
    bot_z_diff = (z2-h2/2)-(z0-h0/2)
    y_diff     = w2/2-w0/2
    x_diff     = x2-x0

    top_angle  = np.tan(top_z_diff/x_diff)/Units.deg
    bot_angle  = np.tan(-bot_z_diff/x_diff)/Units.deg
    side_angle = np.tan(y_diff/x_diff)/Units.deg

    vsp.SetParmVal(b_id,"TBSym","XSec_"+str(i+1),0)
    vsp.SetParmVal(b_id,"TopLAngle","XSec_"+str(i+1),top_angle)
    vsp.SetParmVal(b_id,"TopLStrength","XSec_"+str(i+1),0.75)
    vsp.SetParmVal(b_id,"BottomLAngle","XSec_"+str(i+1),bot_angle)
    vsp.SetParmVal(b_id,"BottomLStrength","XSec_"+str(i+1),0.75)   
    vsp.SetParmVal(b_id,"RightLAngle","XSec_"+str(i+1),side_angle)
    vsp.SetParmVal(b_id,"RightLStrength","XSec_"+str(i+1),0.75)   

    return  

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_boom_fineness
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_boom_fineness(boom, x_locs, eff_diams, eff_diam_gradients_fwd):
    """This computes boom finenesses for nose and tail.

    Assumptions:
    Written for OpenVSP 3.16.1

    Source:
    N/A

    Inputs:
    0. Pre-loaded VSP vehicle in memory, via import_vsp_vehicle.
    1. RCAIDE boom [object].
    2. Array of x_locations of boom segments. (length = L)
    3. Array of effective diameters of boom segments. (length = L)
    4. Array of effective diameter gradients from nose to tail. (length = L-1)

    Outputs:
    Writes fineness values to RCAIDE boom, returns boom.

    Properties Used:
    N/A
    """

    segment_list       = list(boom.segments.keys())
    
    # Compute nose fineness.    
    x_locs    = np.array(x_locs)					# Make numpy arrays.
    eff_diams = np.array(eff_diams)
    min_val   = np.min(eff_diam_gradients_fwd[x_locs[:-1]<=0.5])	# Computes smallest eff_diam gradient value in front 50% of boom.
    x_loc     = x_locs[:-1][eff_diam_gradients_fwd==min_val][0]		# Determines x-location of the first instance of that value (if gradient=0, Segments[segment_list[0]]ost x-loc).
    boom.lengths.nose  = (x_loc-boom.segments[segment_list[0]].percent_x_location)*boom.lengths.total	# Subtracts first segment x-loc in case not at global origin.
    boom.fineness.nose = boom.lengths.nose/(eff_diams[x_locs==x_loc][0])

    # Compute tail fineness.
    x_locs_tail		    = x_locs>=0.5						# Searches aft 50% of boom.
    eff_diam_gradients_fwd_tail = eff_diam_gradients_fwd[x_locs_tail[1:]]			# Smaller array of tail gradients.
    min_val 		    = np.min(-eff_diam_gradients_fwd_tail)			# Computes min gradient, where boom tapers (minus sign makes positive).
    x_loc = x_locs[np.hstack([False,-eff_diam_gradients_fwd==min_val])][-1]			# Saves aft-most value (useful for straight boom with multiple zero gradients.)
    boom.lengths.tail       = (1.-x_loc)*boom.lengths.total
    boom.fineness.tail      = boom.lengths.tail/(eff_diams[x_locs==x_loc][0])	# Minus sign converts tail fineness to positive value.

    return boom

# ---------------------------------------------------------------------------------------------------------------------- 
# get_boom_height
# ---------------------------------------------------------------------------------------------------------------------- 
def get_boom_height(boom, location):
    """This linearly estimates boom height at any percentage point (0,100) along boom length.

    Assumptions:
    Written for OpenVSP 3.16.1

    Source:
    N/A

    Inputs:
    0. Pre-loaded VSP vehicle in memory, via import_vsp_vehicle.
    1. RCAIDE boom [object], containing boom.vsp_data.xsec_num in its data structure.
    2. boom percentage point [float].

    Outputs:
    height [m]

    Properties Used:
    N/A
    """

    segment_list   = list(boom.segments.keys())       
    for jj in range(1, boom.vsp_data.xsec_num):		# Begin at second section, working toward tail.
        if boom.segments[segment_list[jj]].percent_x_location>=location and boom.segments[segment_list[jj-1]].percent_x_location<location:  
            # Find two sections on either side (or including) the desired boom length percentage.
            a        = boom.segments[segment_list[jj]].percent_x_location							
            b        = boom.segments[segment_list[jj-1]].percent_x_location
            a_height = boom.segments[segment_list[jj]].height		# Linear approximation.
            b_height = boom.segments[segment_list[jj-1]].height
            slope    = (a_height - b_height)/(a-b)
            height   = ((location-b)*(slope)) + (b_height)	
            break
    return height

# ---------------------------------------------------------------------------------------------------------------------- 
# find_fuse_u_coordinate
# ---------------------------------------------------------------------------------------------------------------------- 
def find_fuse_u_coordinate(x_target,b_id,fuel_tank_tag):
    """Determines the u coordinate of an OpenVSP boom that matches an x coordinate

    Assumptions:
    boom is aligned with the x axis

    Source:
    N/A

    Inputs:
    x_target      [m]
    b_id       <str>
    fuel_tank_tag <str>

    Outputs:
    u_current     [-] u coordinate for the requests x position

    Properties Used:
    N/A
    """     
    tol   = 1e-3
    diff  = 1000    
    u_min = 0
    u_max = 1    
    while np.abs(diff) > tol:
        u_current = (u_max+u_min)/2
        probe_id = vsp.AddProbe(b_id,0,u_current,0,fuel_tank_tag+'_probe')
        vsp.Update()
        x_id  = vsp.FindParm(probe_id,'X','Measure')
        x_pos = vsp.GetParmVal(x_id) 
        diff = x_target-x_pos
        if diff > 0:
            u_min = u_current
        else:
            u_max = u_current
        vsp.DelProbe(probe_id)
    return u_current




# RCAIDE/Library/Missions/Segments/Climb/Constant_Mach_Linear_Altitude.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE  
from RCAIDE.Library.Mission.Common.Update.atmosphere import atmosphere

# Package imports  
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Constrant dynamic pressure and constant rate of climb

    Source:
    N/A

    Inputs:
    segment.mach_number                         [unitless]
    segment.dynamic_pressure                    [pascals]
    segment.altitude_start                      [meters]
    segment.altitude_end                        [meters]
    segment.distance                            [meters]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """        
    
    # unpack
    alt0       = segment.altitude_start 
    altf       = segment.altitude_end
    xf         = segment.distance
    mach       = segment.mach_number
    beta       = segment.sideslip_angle
    conditions = segment.state.conditions  
    t_nondim   = segment.state.numerics.dimensionless.control_points
    
    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('altitude not set')
        alt0 = -1.0 * segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
        
    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0      
    segment.state.conditions.freestream.altitude[:,0] = alt[:,0]

    # check for initial velocity
    if mach is None: 
        if not segment.state.initials: raise AttributeError('mach not set')
        air_speed = np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1])   
    else:
        
        # Update freestream to get speed of sound
        atmosphere(segment)
        a          = conditions.freestream.speed_of_sound    
    
        # compute speed, constant with constant altitude
        air_speed    = mach * a   
        
    climb_angle  = np.arctan((altf-alt0)/xf)
    v_x          = np.cos(beta)*np.cos(climb_angle)*air_speed
    v_y          = np.sin(beta)*np.cos(climb_angle)*air_speed
    v_z          = np.sin(climb_angle)*air_speed 
    t_nondim     = segment.state.numerics.dimensionless.control_points
    
    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0    
    
    # pack
    conditions.freestream.altitude[:,0]             = alt[:,0]
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down 
    conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    conditions.frames.inertial.velocity_vector[:,2] = -v_z[:,0]      


# RCAIDE/Library/Methods/Thermal_Management/Batteries/Liquid_Cooled_Wavy_Channel/wavy_channel_sizing_setup.py


# Created: Apr 2024 S. Shekar 
#
# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
# RCAIDE Imports  
from RCAIDE.Framework.Analyses.Process   import Process 
from RCAIDE.Library.Components.Thermal_Management.Accessories import Pump 

# Python package imports  
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  Sizing Setup for Wavy Channel
# ----------------------------------------------------------------------------------------------------------------------
def wavy_channel_sizing_setup():
    """ Creates a process to modify and analyze the heat acquistion system
          
          Inputs:  
             None
             
          Outputs:   
             procedure - optimization methodology                                       
              
          Assumptions: 
             N/A 
        
          Source:
             None
    """            

    # size the base config
    procedure = Process()

    # modify battery thermal management system
    procedure.modify_HAS = modify_wavy_channel_HAS

    # post process the results
    procedure.post_process = post_process

    return procedure


def modify_wavy_channel_HAS(nexus): 
    """ Modifies geometry of Wavy Channel  
          
          Inputs:  
             nexus     - RCAIDE optmization framework with Wavy Channel geometry data structure [None]
              
          Outputs:   
             procedure - optimization methodology                                       
              
          Assumptions: 
             N/A 
        
          Source:
             None
    """        
    battery_list  = list(nexus.hrs_configurations.optimized.networks.electric.busses.bus.battery_modules.keys())
    battery       = nexus.hrs_configurations.optimized.networks.electric.busses.bus.battery_modules[battery_list[0]]
    has_opt       = nexus.hrs_configurations.optimized.networks.electric.coolant_lines.coolant_line.battery_modules[battery.tag].thermal_management_system.heat_acquisition_system

    # ------------------------------------------------------------------------------------------------------------------------
    # Unpack paramters  
    # ------------------------------------------------------------------------------------------------------------------------
    
    # Coolant 
    m_coolant = has_opt.coolant_flow_rate 
    coolant   = has_opt.coolant
    T_i       = has_opt.coolant_inlet_temperature  
    
    # Battery 
    d_cell    = battery.cell.diameter                    
    h_cell    = battery.cell.height                      
    A_cell    = np.pi*d_cell*h_cell    
    T_bat     = has_opt.design_battery_operating_temperature
    Q_module  = has_opt.design_heat_removed 
    
    # Channel 
    channel          = has_opt.channel
    b                = has_opt.channel_side_thickness                
    d                = has_opt.channel_width                         
    theta            = has_opt.channel_contact_angle    
    c                = battery.cell.height 
    a                = has_opt.channel_top_thickness    
    AR               = d/c    
    channel_density  = channel.density
    k_chan           = channel.thermal_conductivity  

    # Pump
    n_pump           = 0.7 # Replce with Pump Class


    # Considering a lumped mass model  
    N_cells = battery.electrical_configuration.parallel*battery.electrical_configuration.series
    
    # Contact Surface area of the channel 
    A_chan = 2*N_cells*(theta)*A_cell 
    
    # Update Spacing of the battery pack
    new_normal_spacing  = 2 * (d_cell + d) * np.sin(theta/2)
    new_parllel_spacing  = (d_cell + d) * np.cos(theta/2)
    
    #Length of Channel   
    L_chan  = (battery.geometrtic_configuration.normal_count*new_normal_spacing)*battery.geometrtic_configuration.parallel_count
    L_extra  = 3*d_cell
    L_chan   = (N_cells*d_cell)+L_extra 

    # Hydraulic diameter    
    dh   = (4*c*d)/(2*(c+d))   

    # Thermophysical Properties of Coolant 
    rho  = coolant.compute_density(T_i)
    mu   = coolant.compute_absolute_viscosity(T_i)  
    cp   = coolant.compute_cp(T_i)
    Pr   = coolant.compute_prandtl_number(T_i) 
    k    = coolant.compute_thermal_conductivity(T_i)   

    #calculate the velocity of the fluid in the channel 
    v=rho*c*d*m_coolant

    # calculate the Reynolds Number 
    Re=(rho*dh*v)/mu

    # fanning friction factor (eq 32),  Nusselt Number (eq 12)   
    if Re< 2300:
        f= 24*(1-(1.3553*AR)+(1.9467*(AR**2))-(1.7012*(AR**3))+(0.9564*(AR**4))-(0.2537*(AR**5)))/Re
        Nu = 8.235*(1-(2.0421*AR)+(3.0853*(AR**2))-(2.4765*(AR**3))+(1.0578*(AR**4))-(0.1861*(AR**5)))   
    elif Re>=2300:
        f= (0.0791*(Re**(-0.25)))*(1.8075-0.1125*AR)
        Nu = ((f/2)*(Re-1000)*Pr)/(1+(12.7*((f/2)**0.5)*(Pr**(2/3)-1)))   

    # Calculate the pressure drop in the channel 
    dp     = 2*f*rho*v*v*L_chan/dh	

    # Calculate the Power consumed
    Power   = Pump.compute_power_consumed(dp, rho, m_coolant, n_pump) 

    # Mass calculations - Channel  
    rho_line        = channel_density*(2*a*((2*b)+d)+(2*b*c))
    mass_channel    = rho_line*L_chan

    # Mass calculations - Liquid 
    mass_liquid  = rho*c*d 
    total_mass   = mass_channel+mass_liquid 

    # heat transfer coefficient of the channeled coolant (eq 11)
    h = k*Nu/dh

    # Overall Heat Transfer Coeffipackcient from battery surface to the coolant fluid (eq 10)
    U_total = 1/((1/h)+(b/k_chan))

    # Calculate NTU
    NTU = U_total*A_chan/(m_coolant*cp)
    
    # Effectiveness of the Channel 
    eff_HAS = 1 - np.exp(-NTU)

    # Calculate Outlet Temparture To ( eq 8)
    T_o = ((T_bat-T_i)*(1-np.exp(-NTU)))+T_i

    # Calculate the Log mean temperature 
    T_lm = ((T_bat-T_i)-(T_bat-T_o))/(np.log((T_bat-T_i)/(T_bat-T_o)))

    # Calculated Heat Convected 
    Q_convec = U_total*A_chan*T_lm*eff_HAS   

    # line level properties for optmization 
    has_opt.mass_properties.mass       = total_mass
    has_opt.design_power_draw          = Power
    has_opt.heat_removed               = Q_convec 
    has_opt.heat_generated             = Q_module 
    has_opt.coolant_outlet_temperature = T_o
    has_opt.coolant_pressure_drop      = dp  
    has_opt.surface_area_channel       = A_chan
    has_opt.heat_transfer_efficiency   = eff_HAS
    has_opt.battery_parllel_spacing     = new_parllel_spacing
    has_opt.battery_series_spacing      = new_normal_spacing

    return nexus     
# ----------------------------------------------------------------------
#   Post Process Results to give back to the optimizer
# ----------------------------------------------------------------------   
def post_process(nexus):
    """ Post processes the outputs obtained from the optimizer  
          
          Inputs:  
             nexus     - RCAIDE optmization framework with Wavy Channel geometry data structure [None]
              
          Outputs:   
             nexus     - Objective Function calculation 
              
              
          Assumptions: 
            Factors were applied to mass and power to scale it for the optimizer in lines 218, 219.
        
          Source:
             None
    """            
    
    summary              = nexus.summary  
    battery_list         = list(nexus.hrs_configurations.optimized.networks.electric.coolant_lines.coolant_line.battery_modules.keys())
    battery              = nexus.hrs_configurations.optimized.networks.electric.coolant_lines.coolant_line.battery_modules[battery_list[0]]
    has_opt              = battery.thermal_management_system.heat_acquisition_system 
    Q_line_rem           = has_opt.heat_removed
    Q_line_gen           = has_opt.heat_generated  
    new_normal_spacing   = has_opt.battery_series_spacing   
    new_parallel_spacing = has_opt.battery_parllel_spacing
    channel_width        = has_opt.channel_width        
    channel_thickness    = has_opt.channel_side_thickness 

    # calculate objective 
    spacing                      = new_normal_spacing + new_parallel_spacing
    spacing                      = spacing*100
    total_mass                   = has_opt.mass_properties.mass/100
    Power                        = has_opt.design_power_draw * 10
    summary.mass_power_objective =  (spacing**2+Power**2 + total_mass**2)**(0.5)  

    # calculate heat constraint  
    summary.heat_energy_constraint  = Q_line_rem - Q_line_gen  
    summary.thickness_constraint    = channel_width-(2*channel_thickness)

    return nexus     


# RCAIDE/Framework/Analyses/Aerodynamics/Aerodynamics.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE imports   
from RCAIDE.Framework.Core     import Data
from RCAIDE.Framework.Analyses import Analysis  

# package imports 
import numpy as np

# ----------------------------------------------------------------------------------------------------------------------
#  Aerodynamics
# ---------------------------------------------------------------------------------------------------------------------- 
class Aerodynamics(Analysis):
    """This is the base class for aerodynamics analyses. It contains functions
    that are built into the default class.
    
    Assumptions:
    None
    
    Source:
    N/A
    """
    def __defaults__(self):
        """This sets the default values and methods for the analysis.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """           
        self.tag                               = 'aerodynamics'  
        self.vehicle                           = Data()
        self.settings                          = Data()
        self.settings.maximum_lift_coefficient = np.inf
        
        
    def evaluate(self,state):
        """The default evaluate function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        results   <Results class> (empty)

        Properties Used:
        N/A
        """           
        
        results = Data()
        
        return results
    
    def post_process(self):
        """The default post processing function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Outputs:
        None

        Properties Used:
        N/A
        """         
        
        return     


# RCAIDE/Methods/Energy/Propulsors/design_propeller.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import RCAIDE
from RCAIDE.Framework.Core   import interp2d
from RCAIDE.Library.Methods.Geometry.Airfoil    import compute_airfoil_properties, compute_naca_4series, import_airfoil_geometry

# package imports 
import numpy as np
import scipy as sp 
from scipy.optimize import root 

# ----------------------------------------------------------------------------------------------------------------------  
#  Design Propeller
# ----------------------------------------------------------------------------------------------------------------------   
def design_propeller(prop,number_of_stations=20):
    """ Optimizes propeller chord and twist given input parameters.
          
          Inputs:
          Either design power or thrust
          prop_attributes.
            hub radius                       [m]
            tip radius                       [m]
            rotation rate                    [rad/s]
            freestream velocity              [m/s]
            number of blades               
            number of stations
            design lift coefficient
            design_altitude                  [m]
            airfoil data
            
          Outputs:
          Twist distribution                 [array of radians]
          Chord distribution                 [array of meters]
              
          Assumptions/ Source:
          Based on Design of Optimum Propellers by Adkins and Liebeck
          
    """
    # Unpack
    N            = number_of_stations       # this number determines the discretization of the propeller into stations
    B            = prop.number_of_blades
    R            = prop.tip_radius
    Rh           = prop.hub_radius
    omega        = prop.cruise.design_angular_velocity    # Rotation Rate in rad/s
    V            = prop.cruise.design_freestream_velocity # Freestream Velocity
    Cl           = prop.cruise.design_Cl                  # Design Lift Coefficient
    alt          = prop.cruise.design_altitude
    Thrust       = prop.cruise.design_thrust
    Power        = prop.cruise.design_power
    airfoils     = prop.airfoils 
    a_loc        = prop.airfoil_polar_stations
    
    if (Thrust == None) and (Power== None):
        raise AssertionError('Specify either design thrust or design power!')
    
    elif (Thrust!= None) and (Power!= None):
        raise AssertionError('Specify either design thrust or design power!') 
    
    if V == 0.0:
        V = 1E-6 
        
    # Calculate atmospheric properties
    atmosphere     = RCAIDE.Framework.Analyses.Atmospheric.US_Standard_1976()
    atmo_data      = atmosphere.compute_values(alt) 
    T              = atmo_data.temperature[0]
    rho            = atmo_data.density[0]
    speed_of_sound = atmo_data.speed_of_sound[0]
    mu             = atmo_data.dynamic_viscosity[0]
    nu             = mu/rho
    
    # Nondimensional thrust
    if (Thrust!= None) and (Power == None):
        Tc = 2.*Thrust/(rho*(V*V)*np.pi*(R*R))     
        Pc = 0.0 
    
    elif (Thrust== None) and (Power != None):
        Tc = 0.0   
        Pc = 2.*Power/(rho*(V*V*V)*np.pi*(R*R))  
    
    tol   = 1e-10 # Convergence tolerance

    # Step 1, assume a zeta
    zeta = 0.1 # Assume to be small initially
    
    # Step 2, determine F and phi at each blade station 
    chi0    = Rh/R # Where the propeller blade actually starts
    chi     = np.linspace(chi0,1,N+1) # Vector of nondimensional radii
    chi     = chi[0:N]
    lamda   = V/(omega*R)             # Speed ratio
    r       = chi*R                   # Radial coordinate
    x       = omega*r/V               # Nondimensional distance
    diff    = 1.0                     # Difference between zetas
    n       = omega/(2*np.pi)         # Cycles per second
    D       = 2.*R  
    c       = 0.2 * np.ones_like(chi)
    
    # if user defines airfoil, check dimension of stations
    num_airfoils = len(airfoils.keys())
    if num_airfoils>0:
        if len(a_loc) != N:
            raise AssertionError('\nDimension of airfoil sections must be equal to number of stations on propeller') 
        
        for _,airfoil in enumerate(airfoils):  
            if airfoil.geometry == None: # first, if airfoil geometry data not defined, import from geoemtry files
                if type(airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil: # check if naca 4 series of airfoil from datafile
                    airfoil.geometry = compute_naca_4series(airfoil.NACA_4_Series_code,airfoil.number_of_points)
                else:
                    airfoil.geometry = import_airfoil_geometry(airfoil.coordinate_file,airfoil.number_of_points) 
    
            if airfoil.polars == None: # compute airfoil polars for airfoils
                airfoil.polars = compute_airfoil_properties(airfoil.geometry, airfoil_polar_files= airfoil.polar_files) 
    else:
        print('\nDefaulting to scaled DAE51') 
        
    while diff>tol:      
        # assign chord distribution
        prop.chord_distribution = c  
                         
        #Things that need a loop
        Tcnew   = Tc     
        tanphit = lamda*(1.+zeta/2.)   # Tangent of the flow angle at the tip
        phit    = np.arctan(tanphit)   # Flow angle at the tip
        tanphi  = tanphit/chi          # Flow angle at every station
        f       = (B/2.)*(1.-chi)/np.sin(phit) 
        F       = (2./np.pi)*np.arccos(np.exp(-f)) #Prandtl momentum loss factor
        phi     = np.arctan(tanphi)    # Flow angle at every station
        
        #Step 3, determine the product Wc, and RE
        G       = F*x*np.cos(phi)*np.sin(phi) #Circulation function
        Wc      = 4.*np.pi*lamda*G*V*R*zeta/(Cl*B)
        Ma      = Wc/speed_of_sound
        RE      = Wc/nu

        if num_airfoils>0:
            # assign initial values 
            alpha0   = np.ones(N)*0.05
            
            # solve for optimal alpha to meet design Cl target
            sol      = root(objective, x0 = alpha0 , args=(airfoils,a_loc,RE,Cl,N))
            alpha    = sol.x
            
            # query surrogate for sectional Cls at stations 
            Cdval    = np.zeros_like(RE) 
            for j,airfoil in enumerate(airfoils):                   
                pd          = airfoil.polars
                Cdval_af    = interp2d(RE,alpha,pd.reynolds_numbers, pd.angle_of_attacks, pd.drag_coefficients)
                locs        = np.where(np.array(a_loc) == j )
                Cdval[locs] = Cdval_af[locs]    
                
        else:    
            Cdval   = (0.108*(Cl**4)-0.2612*(Cl**3)+0.181*(Cl**2)-0.0139*Cl+0.0278)*((50000./RE)**0.2)
            alpha   = Cl/(2.*np.pi)
            
        #More Cd scaling from Mach from AA241ab notes for turbulent skin friction
        Tw_Tinf = 1. + 1.78*(Ma**2)
        Tp_Tinf = 1. + 0.035*(Ma**2) + 0.45*(Tw_Tinf-1.)
        Tp      = Tp_Tinf*T
        Rp_Rinf = (Tp_Tinf**2.5)*(Tp+110.4)/(T+110.4) 
        Cd      = ((1/Tp_Tinf)*(1/Rp_Rinf)**0.2)*Cdval
        
        #Step 5, change Cl and repeat steps 3 and 4 until epsilon is minimized 
        epsilon = Cd/Cl  
        
        #Step 6, determine a and a', and W 
        a       = (zeta/2.)*(np.cos(phi)**2.)*(1.-epsilon*np.tan(phi)) 
        W       = V*(1.+a)/np.sin(phi)
        
        #Step 7, compute the chord length and blade twist angle  
        c       = Wc/W
        beta    = alpha + phi # Blade twist angle
    
        #Step 8, determine 4 derivatives in I and J 
        Iprime1 = 4.*chi*G*(1.-epsilon*np.tan(phi))
        Iprime2 = lamda*(Iprime1/(2.*chi))*(1.+epsilon/np.tan(phi)
                                            )*np.sin(phi)*np.cos(phi)
        Jprime1 = 4.*chi*G*(1.+epsilon/np.tan(phi))
        Jprime2 = (Jprime1/2.)*(1.-epsilon*np.tan(phi))*(np.cos(phi)**2.) 
        dchi    = (chi[1]-chi[0])*np.ones_like(Jprime1)
        
        #Integrate derivatives from chi=chi0 to chi=1 
        I1      = np.dot(Iprime1,dchi)
        I2      = np.dot(Iprime2,dchi)
        J1      = np.dot(Jprime1,dchi)
        J2      = np.dot(Jprime2,dchi)        

        #Step 9, determine zeta and and Pc or zeta and Tc 
        if (Pc==0.)&(Tc!=0.): 
            #First Case, Thrust is given
            #Check to see if Tc is feasible, otherwise try a reasonable number
            if Tcnew>=I2*(I1/(2.*I2))**2.:
                Tcnew = I2*(I1/(2.*I2))**2.
            zetan    = (I1/(2.*I2)) - ((I1/(2.*I2))**2.-Tcnew/I2)**0.5

        elif (Pc!=0.)&(Tc==0.): 
            #Second Case, Thrust is given
            zetan    = -(J1/(J2*2.)) + ((J1/(J2*2.))**2.+Pc/J2)**0.5 
    
        #Step 10, repeat starting at step 2 with the new zeta
        diff = abs(zeta-zetan)
        
        zeta = zetan
    
    # Step 11, determine propeller efficiency etc...
    if (Pc==0.)&(Tc!=0.): 
        if Tcnew>=I2*(I1/(2.*I2))**2.:
            Tcnew = I2*(I1/(2.*I2))**2.
            print('Tc infeasible, reset to:')
            print(Tcnew)        
        #First Case, Thrust is given
        zeta    = (I1/(2.*I2)) - ((I1/(2.*I2))**2.-Tcnew/I2)**0.5
        Pc      = J1*zeta + J2*(zeta**2.)
        Tc      = I1*zeta - I2*(zeta**2.)
        
    elif (Pc!=0.)&(Tc==0.): 
        #Second Case, Thrust is given
        zeta    = -(J1/(2.*J2)) + ((J1/(2.*J2))**2.+Pc/J2)**0.5
        Tc      = I1*zeta - I2*(zeta**2.)
        Pc      = J1*zeta + J2*(zeta**2.) 
        
    # Calculate mid-chord alignment angle, MCA
    # This is the distance from the mid chord to the line axis out of the center of the blade
    # In this case the 1/4 chords are all aligned 
    MCA    = c/4. - c[0]/4.
    
    Thrust = Tc*rho*(V**2)*np.pi*(R**2)/2
    Power  = Pc*rho*(V**3)*np.pi*(R**2)/2 
    Ct     = Thrust/(rho*(n*n)*(D*D*D*D))
    Cp     = Power/(rho*(n*n*n)*(D*D*D*D*D))  
    
    # compute max thickness distribution  
    t_max  = np.zeros(N)    
    t_c    = np.zeros(N)   
    if num_airfoils>0:
        for j,airfoil in enumerate(airfoils): 
            a_geo         = airfoil.geometry
            locs          = np.where(np.array(a_loc) == j )
            t_max[locs]   = a_geo.max_thickness*c[locs] 
            t_c[locs]     = a_geo.thickness_to_chord 
    else:     
        c_blade = np.repeat(np.atleast_2d(np.linspace(0,1,N)),N, axis = 0)* np.repeat(np.atleast_2d(c).T,N, axis = 1)
        t       = (5*c_blade)*(0.2969*np.sqrt(c_blade) - 0.1260*c_blade - 0.3516*(c_blade**2) + 0.2843*(c_blade**3) - 0.1015*(c_blade**4)) # local thickness distribution
        t_max   = np.max(t,axis = 1) 
        t_c     = np.max(t,axis = 1) /c  
            
    # Nondimensional thrust
    if prop.cruise.design_power == None: 
        prop.cruise.design_power = Power[0]        
    elif prop.cruise.design_thrust == None: 
        prop.cruise.design_thrust = Thrust[0]       
    
    # blade solidity
    r          = chi*R                     
    blade_area = sp.integrate.cumulative_trapezoid(B*c, r-r[0])
    sigma      = blade_area[-1]/(np.pi*R**2)   
    
    prop.cruise.design_torque                   = Power[0]/omega
    prop.max_thickness_distribution             = t_max
    prop.twist_distribution                     = beta
    prop.chord_distribution                     = c
    prop.radius_distribution                    = r
    prop.number_of_blades                       = int(B)
    prop.cruise.design_power_coefficient        = Cp
    prop.cruise.design_thrust_coefficient       = Ct
    prop.mid_chord_alignment                    = MCA
    prop.thickness_to_chord                     = t_c
    prop.blade_solidity                         = sigma

    return prop

    
def objective(x,airfoils,a_loc,RE,Cl,N):
    # query surrogate for sectional Cls at stations 
    Cl_vals          = np.zeros(N)       
    for j,airfoil in enumerate(airfoils): 
        pd            = airfoil.polars
        Cl_af         = interp2d(RE,x,pd.reynolds_numbers, pd.angle_of_attacks, pd.lift_coefficients)
        locs          = np.where(np.array(a_loc) == j )
        Cl_vals[locs] = Cl_af[locs] 
        
    # compute Cl residual    
    Cl_residuals = Cl_vals - Cl 
    return  Cl_residuals 



# RCAIDE/Library/Methods/Propulsors/Converters/Combustor/compute_supersonic_combustion.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------     

# package imports
import numpy as np 

# ---------------------------------------------------------------------------------------------------------------------- 
# calculate_power_from_throttle
# ----------------------------------------------------------------------------------------------------------------------    
def compute_supersonic_combustion(combustor,combustor_conditions,conditions): 
    """ This function computes the output values for supersonic combustion (Scramjet).
    This will be done using stream thrust analysis. The following properties are computed. 
    combustor_conditions.outputs. 
       stagnation_temperature               (numpy.ndarray) : [K] 
       stagnation_pressure                  (numpy.ndarray) : [Pa] 
       stagnation_enthalpy                  (numpy.ndarray) : [J/kg] 
       fuel_to_air_ratio                    (numpy.ndarray) : [unitless] 
       static_temperature                   (numpy.ndarray) : [K] 
       static_pressure                      (numpy.ndarray) : [Pa] 
       velocity                             (numpy.ndarray) : [m/s] 
       mach_number                          (numpy.ndarray) : [unitless]   
    
    
    Assumptions: 
        1. Constant Pressure Combustion      
        2. Flow is in axial direction at all times 
        3. Flow properities at exit are 1-Da averages 

    Source: 
        Heiser, William H., Pratt, D. T., Daley, D. H., and Unmeel, B. M., 
        "Hypersonic Airbreathing Propulsion", 1994 
        Chapter 4 - pgs. 175-180
    
    Args: 
        conditions.freestream. 
           isentropic_expansion_factor              (numpy.ndarray): [unitless] 
           specific_heat_at_constant_pressure       (numpy.ndarray): [J/(kg K)] 
           temperature                              (numpy.ndarray): [K] 
           stagnation_temperature                   (numpy.ndarray): [K]
           universal_gas_constant                   (numpy.ndarray): [J/(kg K)]  
        combustor_conditions.inputs.     
           stagnation_temperature                   (numpy.ndarray): [K] 
           stagnation_pressure                      (numpy.ndarray): [Pa] 
           inlet_nozzle                             (numpy.ndarray): [unitless] 
          combustor.fuel_data.specific_energy       (numpy.ndarray): [J/kg] 
          combustor.efficiency                              (float): [unitless]
          combustor.axial_fuel_velocity_ratio               (float): [unitless] 
          combustor.fuel_velocity_ratio                     (float): [unitless] 
          combustor.burner_drag_coefficient                 (float): [unitless] 
          combustor.temperature_reference                   (float): [K] 
          combustor.absolute_sensible_enthalpy              (float): [J/kg] 
          combustor.specific_heat_constant_pressure         (float): [J/(kg K)] 

    Returns:  
      """  
    # unpacking conditions 
    R      = conditions.freestream.gas_specific_constant 
    Tref   = conditions.freestream.temperature
    
    # unpacking properties of combustor 
    nozzle  = combustor_conditions.inputs.inlet_nozzle 
    Pt_in   = combustor_conditions.inputs.stagnation_pressure 
    Cp_c    = nozzle.specific_heat_at_constant_pressure 
    htf     = combustor.fuel_data.specific_energy 
    eta_b   = combustor.efficiency 
    Vfx_V3  = combustor.axial_fuel_velocity_ratio 
    Vf_V3   = combustor.fuel_velocity_ratio 
    Cfb     = combustor.burner_drag_coefficient 
    hf      = combustor.absolute_sensible_enthalpy 
    phi     = combustor.fuel_equivalency_ratio
    
    # compute gamma and Cp over burner 
    Cpb     = 1.45 * Cp_c         
    gamma_b = (Cpb/R)/(Cpb/R-1.)  
    
    # unpack nozzle input values 
    T_in = nozzle.static_temperature 
    V_in = nozzle.velocity 
    P_in = nozzle.static_pressure 
    
    # setting stoichiometric fuel-to-air   
    f    = phi*combustor.fuel_data.stoichiometric_fuel_to_air
    
    # compute output velocity, mach and temperature 
    V_out  = V_in*(((1.+f*Vfx_V3)/(1.+f))-(Cfb/(2.*(1.+f)))) 
    T_out  = ((T_in/(1.+f))*(1.+(1./(Cpb*T_in ))*(eta_b*f*htf+f*hf+f*Cpb*Tref+(1.+f*Vf_V3*Vf_V3)*V_in*V_in/2.))) - V_out*V_out/(2.*Cpb) 
    M_out  = V_out/(np.sqrt(gamma_b*R*T_out)) 
    Tt_out = T_out*(1.+(gamma_b-1.)/2.)*M_out*M_out
    
    # compute the exity static and stagnation conditions 
    ht_out = Cpb*Tt_out 
    P_out  = P_in 
    Pt_out = Pt_in*((((gamma_b+1.)*(M_out**2.))/((gamma_b-1.)*M_out**2.+2.))**(gamma_b/(gamma_b-1.)))*((gamma_b+1.)/(2.*gamma_b*M_out**2.-(gamma_b-1.)))**(1./(gamma_b-1.))  
    
    # pack results   
    combustor_conditions.outputs.stagnation_temperature          = Tt_out  
    combustor_conditions.outputs.stagnation_pressure             = Pt_out        
    combustor_conditions.outputs.stagnation_enthalpy             = ht_out        
    combustor_conditions.outputs.fuel_to_air_ratio               = f        
    combustor_conditions.outputs.static_temperature              = T_out  
    combustor_conditions.outputs.static_pressure                 = P_out         
    combustor_conditions.outputs.velocity                        = V_out  
    combustor_conditions.outputs.mach_number                     = M_out 
    combustor_conditions.outputs.specific_heat_constant_pressure = Cpb
    combustor_conditions.outputs.isentropic_expansion_factor     = gamma_b
    return 




# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_rotor_performance.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
import pandas as pd 
import plotly.graph_objects as go
from plotly.subplots import make_subplots 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------      
def plot_rotor_performance(rotor,outputs, title=None, show_figure = True,save_figure=False, save_filename='Rotor_Performance', file_type=".png"):
    """Plots a summary of rotor performance 
    
    Assumptions:
    None

    Source: 
    None
    
    Inputs
    rotor   - rotor data structure 
    
    Outputs:
    Plots

    Properties Used:
    N/A
    """
    # unpack 
    r_distribution = outputs.disc_radial_distribution[0, :, 0]
    
    # 2d plots
    fig = make_subplots(rows=2, cols=2)
    
    df1a = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_axial_velocity[0, :, 0])) # label='Axial'
    df1b = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_tangential_velocity[0, :, 0]))  # label='Tangential'
    df2a = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_axial_induced_velocity[0, :, 0])) # label='Axial'
    df2b = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_tangential_induced_velocity[0, :, 0])) # label='Tangential'
    df3  = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_thrust_distribution[0, :, 0]))
    df4  = pd.DataFrame(dict(x=r_distribution, y=outputs.disc_torque_distribution[0, :, 0]))
    
    fig.append_trace(go.Line(df1a, name='Axial', legendgroup='1',showlegend=True), row=1, col=1)
    fig.append_trace(go.Line(df1b, name='Tangential', legendgroup='1',showlegend=True), row=1, col=1)
    fig.append_trace(go.Line(df2a, name='Axial', legendgroup='2',showlegend=True), row=1, col=2)    
    fig.append_trace(go.Line(df2b, name='Tangential', legendgroup='2',showlegend=True), row=1, col=2)     
    fig.append_trace(go.Line(df3, name='Thrust', legendgroup='3',showlegend=False), row=2, col=1)    
    fig.append_trace(go.Line(df4, name='Torque', legendgroup='4',showlegend=False), row=2, col=2)        
    
    fig.update_xaxes(title_text="Radial Station", row=1, col=1)
    fig.update_yaxes(title_text="Velocity", row=1, col=1)
    fig.update_xaxes(title_text="Radial Station", row=1, col=2)
    fig.update_yaxes(title_text="Induced Velocity", row=1, col=2)
    fig.update_xaxes(title_text="Radial Station", row=2, col=1)
    fig.update_yaxes(title_text="Thrust, N", row=2, col=1)
    fig.update_xaxes(title_text="Radial Station", row=2, col=2)
    fig.update_yaxes(title_text="Torque, N-m", row=2, col=2)
    
    fig.update_layout(title_text="Rotor Performance", height=700)
    
    if save_figure:
        fig.write_image(save_filename + '_2D' + file_type) 
    
    if show_figure:
        fig.write_html( save_filename + '.html', auto_open=True)
    return fig 
 


 # RCAIDE/Library/Missions/Segments/Descent/Linear_Mach_Constant_Rate.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------  
#  IMPORT 
# ----------------------------------------------------------------------------------------------------------------------  
# RCAIDE imports  
from RCAIDE.Library.Mission.Common.Update.atmosphere import atmosphere

# pacakge imports 
import numpy as np

# ---------------------------------------------------------------------------------------------------------------------- 
#  Initialize Conditions
# ----------------------------------------------------------------------------------------------------------------------  
def initialize_conditions(segment):
    """Sets the specified conditions which are given for the segment type.

    Assumptions:
    Change mach linearly through the descent with constant descent rate

    Source:
    N/A

    Inputs:
    segment.descent_rate                        [meters/second]
    segment.altitude_start                      [meters]
    segment.altitude_end                        [meters]
    segment.mach_number_start                          [unitless]
    segment.mach_number_end                            [unitless]
    segment.air_speed                           [meters/second]
    state.numerics.dimensionless.control_points [array]

    Outputs:
    conditions.frames.inertial.velocity_vector  [meters/second]
    conditions.frames.inertial.position_vector  [meters]
    conditions.freestream.altitude              [meters]
    conditions.frames.inertial.time             [seconds]

    Properties Used:
    N/A
    """      
    
    # unpack
    descent_rate = segment.descent_rate
    M0           = segment.mach_number_start
    Mf           = segment.mach_number_end
    alt0         = segment.altitude_start 
    altf         = segment.altitude_end 
    beta         = segment.sideslip_angle    
    t_nondim     = segment.state.numerics.dimensionless.control_points
    conditions   = segment.state.conditions  

    # Update freestream to get speed of sound
    atmosphere(segment)
    a          = conditions.freestream.speed_of_sound        
    
    # check for initial altitude
    if alt0 is None:
        if not segment.state.initials: raise AttributeError('initial altitude not set')
        alt0 = -1.0 *segment.state.initials.conditions.frames.inertial.position_vector[-1,2]
    
    # check for initial velocity vector
    if M0 is None:
        if not segment.state.initials: raise AttributeError('initial mach number not set')
        M0  =  np.linalg.norm(segment.state.initials.conditions.frames.inertial.velocity_vector[-1,:])/a[0,:]         
        
    # discretize on altitude
    alt = t_nondim * (altf-alt0) + alt0
    conditions.freestream.altitude[:,0] =  alt[:,0]  # positive altitude 

    # process velocity vector
    mach_number = (Mf-M0)*t_nondim + M0
    v_xy_mag    = mach_number * a
    v_z         = descent_rate # z points down
    v_xy        = np.sqrt( v_xy_mag**2 - v_z**2 ) 
    v_x         = np.cos(beta)*v_xy
    v_y         = np.sin(beta)*v_xy
    
    # pack conditions    
    conditions.frames.inertial.velocity_vector[:,0] = v_x[:,0]
    conditions.frames.inertial.velocity_vector[:,1] = v_y[:,0]
    conditions.frames.inertial.velocity_vector[:,2] = v_z
    conditions.frames.inertial.position_vector[:,2] = -alt[:,0] # z points down
    conditions.freestream.altitude[:,0]             =  alt[:,0] # positive altitude t



# RCAIDE/Methods/Noise/Frequency_Domain_Buildup/Rotor/__init__.py
# 

""" RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
  
from .compute_rotor_noise           import compute_rotor_noise 
from .broadband_noise               import broadband_noise
from .harmonic_noise_line           import harmonic_noise_line
from .harmonic_noise_plane          import harmonic_noise_plane
from .harmonic_noise_point          import harmonic_noise_point          
from .BPM_boundary_layer_properties import BPM_boundary_layer_properties 
from .LBL_VS_broadband_noise        import LBL_VS_broadband_noise       
from .TBL_TE_broadband_noise        import TBL_TE_broadband_noise       
from .TIP_broadband_noise           import TIP_broadband_noise          
from .noise_directivities           import noise_directivities          


## @ingroup Methods-Dynamo_Supply-dynamo_supply_mass_estimation
# dynamo_supply_mass_estimation.py
#
# Created:  Feb 2020,   K. Hamilton - Through New Zealand Ministry of Business Innovation and Employment Research Contract RTVU2004 
# Modified: Feb 2022,   S. Claridge 
# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# import math
import numpy as np
import scipy.integrate as integrate

# ----------------------------------------------------------------------
#  Dynamo_Supply dynamo_supply_mass_estimation
# ----------------------------------------------------------------------    
    
    
    
def dynamo_supply_mass_estimation(HTS_Dynamo_Supply):
    """ Basic mass estimation for HTS Dynamo supply. This supply includes all elements required to create the required shaft power from supplied electricity, i.e. the esc, brushless motor, and gearbox.
    Assumptions:
        Mass scales linearly with power and current

    Source:
        Maxon Motor drivetrains

    Inputs:
        current             [A]
        power_out           [W]

    Outputs:
        mass                [kg]

    Properties Used:
        None
    """

    # unpack
    rated_power     = HTS_Dynamo_Supply.rated_power

    # Estimate mass of motor and gearbox. Source: Maxon EC-max 12V brushless motors under 100W.
    mass_motor      = 0.013 + 0.0046 * rated_power
    mass_gearbox    = 0.0109 + 0.0015 * rated_power

    # Estimate mass of motor driver (ESC). Source: Estimate
    mass_esc        = (5.0 + rated_power/50.0)/1000.0

    # Sum masses to give total mass
    mass            = mass_esc + mass_motor + mass_gearbox

    # Store results
    HTS_Dynamo_Supply.mass_properties.mass       = mass

    # Return results
    return mass


# RCAIDE/Library/Missions/Common/Update/stability.py
# 
# 
# Created:  Jul 2023, M. Clarke 

import numpy as np
import RCAIDE
# ----------------------------------------------------------------------------------------------------------------------
#  Stability
# ---------------------------------------------------------------------------------------------------------------------- 
def stability(segment): 
    """ Updates the stability of the aircraft 
        
        Assumptions:
        If stability model is defined, overwrite the aerodynamics calculations
        
        Inputs:
            None 
                 
        Outputs: 
            None
      
        Properties Used:
        N/A
                    
    """   
    # unpack
    stability_model    = segment.analyses.stability
    conditions         = segment.state.conditions
     
    if stability_model != None:
        
        # check to see if results have already been calculated using the Aero analyses
        if ((type(segment.analyses.stability) == RCAIDE.Framework.Analyses.Stability.Vortex_Lattice_Method) and (type(segment.analyses.aerodynamics) == RCAIDE.Framework.Analyses.Aerodynamics.Vortex_Lattice_Method)) or \
            ((type(segment.analyses.stability) == RCAIDE.Framework.Analyses.Stability.Athena_Vortex_Lattice) and (type(segment.analyses.aerodynamics) == RCAIDE.Framework.Analyses.Aerodynamics.Athena_Vortex_Lattice)):
            pass
        else:
            Sref               = stability_model.vehicle.reference_area
            MAC                = stability_model.vehicle.wings.main_wing.chords.mean_aerodynamic
            span               = stability_model.vehicle.wings.main_wing.spans.projected
            q                  = segment.state.conditions.freestream.dynamic_pressure
            CLmax              = stability_model.settings.maximum_lift_coefficient
    
            _ = stability_model(segment)
     
            # Forces 
            CL = conditions.aerodynamics.coefficients.lift.total
            CD = conditions.aerodynamics.coefficients.drag.total
            CY = conditions.static_stability.coefficients.Y
    
            CL[q<=0.0] = 0.0
            CD[q<=0.0] = 0.0
            CL[CL>CLmax] = CLmax
            CL[CL< -CLmax] = -CLmax
    
            # dimensionalize
            F      = segment.state.ones_row(3) * 0.0
            F[:,2] = ( -CL * q * Sref )[:,0]
            F[:,1] = ( -CY * q * Sref )[:,0]
            F[:,0] = ( -CD * q * Sref )[:,0]
    
            # rewrite aerodynamic CL and CD
            conditions.aerodynamics.coefficients.lift.total  = CL
            conditions.aerodynamics.coefficients.drag.total  = CD
            conditions.frames.wind.force_vector[:,:]   = F[:,:]
    
            # -----------------------------------------------------------------
            # Moments
            # -----------------------------------------------------------------
            CM = conditions.static_stability.coefficients.M
            CL = conditions.static_stability.coefficients.L
            CN = conditions.static_stability.coefficients.N
    
            CM[q<=0.0] = 0.0
    
            # dimensionalize
            M      = segment.state.ones_row(3) * 0.0
            M[:,0] = (CL[:,0] * q[:,0] * Sref * span)
            M[:,1] = (CM[:,0] * q[:,0] * Sref * MAC)
            M[:,2] = (CN[:,0] * q[:,0] * Sref * span)
    
            # pack conditions
            conditions.frames.wind.moment_vector[:,:] = M[:,:]

    return


 # Coolant_use.py
#
# Created:  Feb 2020,   K. Hamilton - Through New Zealand Ministry of Business Innovation and Employment Research Contract RTVU2004 
# Modified: Feb 2022,   S. Claridge 

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

# import math
import numpy as np
import scipy.integrate as integrate

# ----------------------------------------------------------------------
#  Cryogen Coolant_use
# ---------------------------------------------------------------------- 
def compute_cryogen_mass_flow_rate(cryogen,cryogen_temp,equipment_temp,cooling_required,pressure):
    """ Calculate the mass flow rate of cryogen required to maintain cryogenic equipment temperature

    Assumptions:
    constant pressure cryogen system
    perfect mixing and thermal conduction in cryogen
    
    Inputs:
        cryogen
        cryogen_temp                [K]
        equipment_temp              [K]
        cooling_power               [W]
        pressure                    [Pa]
    
    Outputs:
        mdot                        [kg/s]
    
    """
    
    # Method:
    # Given the cryogen pressure find the boiling temperature using the Antoine Equation.
    # T = B/(A-log10(P))-C
    boil_temp = (cryogen.antoine_B / (cryogen.antoine_A - np.log10(pressure))) - cryogen.antoine_C

    # Define cooling type flags and initial values
    Liquid_Cooled       = False
    Evap_Cooled         = False
    Gas_Cooled          = False
    liq_cooling         = 0.0
    gas_cooling         = 0.0
    evap_cooling        = 0.0

    # Calculate cooling type based on boiling temp and required temp
    # Then set the limits of each cooling type based on whether boiling occurs
    if boil_temp < equipment_temp:
        Gas_Cooled      = True
        gas_Ti          = cryogen_temp
        gas_Tf          = equipment_temp
    if boil_temp > cryogen_temp:
        Liquid_Cooled   = True
        liq_Ti          = cryogen_temp
        liq_Tf          = equipment_temp
    if Gas_Cooled and Liquid_Cooled:
        Evap_Cooled     = True
        liq_Tf          = boil_temp
        gas_Ti          = boil_temp

    # Calculate the cooling power per gram of coolant for each of the cooling modes
    # LIQUID COOLING
    # find the area under C_P vs temperature between the temperatures over which the cryogen is a liquid, as this will be the cooling available (in Joules) per gram, aka watts per gram per second.
    if Liquid_Cooled:
        liq_cooling = integrate.quad(lambda t: (cryogen.LCP_C3*t**3. + cryogen.LCP_C2*t**2. + cryogen.LCP_C1*t**1. + cryogen.LCP_C0*t**0.), liq_Ti, liq_Tf)[0]

    # GAS COOLING
    # find the area under C_P vs temperature between the temperatures over which the cryogen is a vapour, as this will be the cooling available (in Joules) per gram, aka watts per gram per second.
    if Gas_Cooled:
        gas_cooling = integrate.quad(lambda t: (cryogen.GCP_C3*t**3. + cryogen.GCP_C2*t**2. + cryogen.GCP_C1*t**1. + cryogen.GCP_C0*t**0.), gas_Ti, gas_Tf)[0]

    # EVAPORATION COOLING
    # Calculate the enthalpy using the polynomial fit to pressure. Enthalpy is in kJ/kg, i.e. J/g.
    if Evap_Cooled:
        evap_cooling = cryogen.H_C3*pressure**3. + cryogen.H_C2*pressure**2. + cryogen.H_C1*pressure**1. + cryogen.H_C0*pressure**0.

    # Sum the components of cooling to give the total cooling power per gram. X1000 to give per kg.
    cooling_power = 1000. * (liq_cooling + evap_cooling + gas_cooling)

    # Divide the cooling power by the per kg cooling power to calculate the coolant mass flow rate    
    mdot = cooling_required/cooling_power       # [kg/s]    

    # Return mass flow rate of the cryogen         
    return mdot


# RCAIDE/Library/Methods/Propulsors/Converters/Compression_Nozzle/compute_supersonic_nozzle_performance.py
# (c) Copyright 2023 Aerospace Research Community LLC
# 
# Created:  Jun 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------     

# package imports
import numpy as np   
from RCAIDE.Library.Methods.Gas_Dynamics.fm_id import fm_id

# ---------------------------------------------------------------------------------------------------------------------- 
# compute_compression_nozzle_performance
# ----------------------------------------------------------------------------------------------------------------------    
def compute_supersonic_nozzle_performance(supersonic_nozzle,s_nozzle_conditions,conditions): 
    """This computes the output values from the input values according to
    equations from the source.
    
    Assumptions:
    Constant polytropic efficiency and pressure ratio
    
    Source:
    https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/
    
    Inputs:
    conditions.freestream.
      isentropic_expansion_factor         [-]
      specific_heat_at_constant_pressure  [J/(kg K)]
      pressure                            [Pa]
      stagnation_pressure                 [Pa]
      stagnation_temperature              [K]
      gas_specific_constant               [J/(kg K)]
      mach_number                         [-]
    self.inputs.
      stagnation_temperature              [K]
      stagnation_pressure                 [Pa]
               
    Outputs:
    s_nozzle_conditions.
      stagnation_temperature              [K]  
      stagnation_pressure                 [Pa]
      stagnation_enthalpy                 [J/kg]
      mach_number                         [-]
      static_temperature                  [K]
      static_enthalpy                     [J/kg]
      velocity                            [m/s]
      static_pressure                     [Pa]
      area_ratio                          [-]
            
    Properties Used:
    self.
      pressure_ratio                      [-]
      polytropic_efficiency               [-]
      pressure_recovery                   [-]
    """           
    
    #unpack the values
    
    #unpack from conditions
    gamma    = conditions.freestream.isentropic_expansion_factor
    Cp       = conditions.freestream.specific_heat_at_constant_pressure
    Po       = conditions.freestream.pressure
    Pto      = conditions.freestream.stagnation_pressure
    Tto      = conditions.freestream.stagnation_temperature
    R        = conditions.freestream.gas_specific_constant
    Mo       = conditions.freestream.mach_number
    
    #unpack from inputs
    Tt_in    = s_nozzle_conditions.inputs.stagnation_temperature
    Pt_in    = s_nozzle_conditions.inputs.stagnation_pressure
     
    pid      = supersonic_nozzle.pressure_ratio
    etapold  = supersonic_nozzle.polytropic_efficiency
    eta_rec  = supersonic_nozzle.pressure_recovery
    
    #Method for computing the nozzle properties
    
    #--Getting the output stagnation quantities
    Pt_out   = Pt_in*pid*eta_rec
    Tt_out   = Tt_in*(pid*eta_rec)**((gamma-1)/(gamma)*etapold)
    ht_out   = Cp*Tt_out
    
    
    #compute the output Mach number, static quantities and the output velocity
    Mach          = np.sqrt((((Pt_out/Po)**((gamma-1)/gamma))-1)*2/(gamma-1))
    
    #Remove check on mach numbers from expansion nozzle
    i_low         = Mach < 10.0
    
    #initializing the Pout array
    P_out         = 1.0 *Mach/Mach
    
    #Computing output pressure and Mach number for the case Mach <1.0
    P_out[i_low]  = Po[i_low]
    Mach[i_low]   = np.sqrt((((Pt_out[i_low]/Po[i_low])**((gamma[i_low]-1.)/gamma[i_low]))-1.)*2./(gamma[i_low]-1.))
    
    #Computing the output temperature,enthalpy, velocity and density
    T_out         = Tt_out/(1.+(gamma-1.)/2.*Mach*Mach)
    h_out         = Cp*T_out
    u_out         = np.sqrt(2.*(ht_out-h_out))
    rho_out       = P_out/(R*T_out)
    
    #Computing the freestream to nozzle area ratio (mainly from thrust computation)
    area_ratio    = (fm_id(Mo,gamma)/fm_id(Mach,gamma)*(1/(Pt_out/Pto))*(np.sqrt(Tt_out/Tto)))
    
    #pack computed quantities into outputs
    s_nozzle_conditions.outputs.stagnation_temperature  = Tt_out
    s_nozzle_conditions.outputs.stagnation_pressure     = Pt_out
    s_nozzle_conditions.outputs.stagnation_enthalpy     = ht_out
    s_nozzle_conditions.outputs.mach_number             = Mach
    s_nozzle_conditions.outputs.static_temperature      = T_out
    s_nozzle_conditions.outputs.density                 = rho_out
    s_nozzle_conditions.outputs.static_enthalpy         = h_out
    s_nozzle_conditions.outputs.velocity                = u_out
    s_nozzle_conditions.outputs.static_pressure         = P_out
    s_nozzle_conditions.outputs.area_ratio              = area_ratio
        
    return 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Common/compute_systems_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ---------------------------------------------------------------------------------------------------------------------- 
from RCAIDE.Framework.Core import Data, Units
import RCAIDE.Library.Components.Wings as Wings 

# ---------------------------------------------------------------------------------------------------------------------- 
# Payload
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_systems_weight(vehicle):
    """ Calculate the weight of the different engine systems on the aircraft

    Assumptions:
        numbers based on FAA regulations and correlations from previous aircraft

    Source:
        http://aerodesign.stanford.edu/aircraftdesign/structures/componentweight.html

   Inputs:
       vehicle.passengers - total number of seats on the aircraft                                     [dimensionless]
       vehicle.systems.control - specifies if the control system is fully power,
                                    partially powered, or not powered                                 [dimensionless]
       wing.areas.reference - area of the horizontal tail                                             [meters**2]
       wing.areas.reference - area of the vertical tail                                               [meters**2]
       vehicle.reference_area - area of the wing                                                      [meters**2]
       vehicle.systems.accessories - determines type of instruments, electronics,
                                        and operating items based on type of vehicle                  [dimensionless]

   Outputs:
       output - a data dictionary with fields:
           W_flight_controls - weight of the flight control system                                               [kilograms]
           W_apu - weight of the apu                                                                      [kilograms]
           W_hyd_pnu - weight of the hydraulics and pneumatics                                            [kilograms]
           W_instruments - weight of the instruments and navigational equipment                           [kilograms]
           W_avionics - weight of the avionics                                                            [kilograms]
           W_opitems - weight of the optional items based on the type of aircraft                         [kilograms]
           W_electrical - weight of the electrical items                                                        [kilograms]
           W_ac - weight of the air conditioning and anti-ice system                                      [kilograms]
           W_furnish - weight of the furnishings in the fuselage                                          [kilograms]

    Properties Used:
        N/A
    """ 
    num_seats   = vehicle.passengers
    ctrl_type   = vehicle.systems.control
    ac_type     = vehicle.systems.accessories
    S_gross_w   = vehicle.reference_area
    sref        = S_gross_w / Units.ft ** 2  # Convert meters squared to ft squared
    s_tail      = 0
    for wing in vehicle.wings:
        if isinstance(wing, Wings.Horizontal_Tail) or isinstance(wing, Wings.Vertical_Tail):
            s_tail += wing.areas.reference
    if s_tail == 0: # assume flight control only on wing, for example on a BWB
        for wing in vehicle.wings:
            if isinstance(wing, Wings.Main_Wing):
                s_tail += wing.areas.reference * 0.01
    area_hv = s_tail / Units.ft ** 2  # Convert meters squared to ft squared
 
    # Flight Controls Group Wt
    if ctrl_type == "fully powered":  # fully powered controls
        flt_ctrl_scaler = 3.5
    elif ctrl_type == "partially powered":  # partially powered controls
        flt_ctrl_scaler = 2.5
    else:
        flt_ctrl_scaler = 1.7  # fully aerodynamic controls
    W_flight_controls = (flt_ctrl_scaler * (area_hv)) * Units.lb

    # APU Group Wt
    if num_seats >= 6.:
        apu_wt = 7.0 * num_seats * Units.lb
    else:
        apu_wt = 0.0 * Units.lb  # no apu if less than 9 seats
    apu_wt = max(apu_wt, 70.)
    
    # Hydraulics & Pneumatics Group Wt
    hyd_pnu_wt = (0.65 * sref) * Units.lb

    # Electrical Group Wt
    W_electrical = (13.0 * num_seats) * Units.lb

    # Furnishings Group Wt
    W_furnish = ((43.7 - 0.037 * min(num_seats, 300.)) * num_seats + 46.0 * num_seats) * Units.lb

    # Environmental Control
    W_air_conditioning = (15.0 * num_seats) * Units.lb

    # Instruments, Electronics, Operating Items based on Type of Vehicle 
    if ac_type == "short-range":  # short-range domestic, austere accomodations
        W_instruments = 800.0 * Units.lb
        W_avionics = 900.0 * Units.lb
    elif ac_type == "medium-range":  # medium-range domestic
        W_instruments = 800.0 * Units.lb
        W_avionics = 900.0 * Units.lb
    elif ac_type == "long-range":  # long-range overwater
        W_instruments = 1200.0 * Units.lb
        W_avionics = 1500.0 * Units.lb
        W_furnish += 23.0 * num_seats * Units.lb  # add aditional seat wt
    elif ac_type == "business":  # business jet
        W_instruments = 100.0 * Units.lb
        W_avionics = 300.0 * Units.lb
    elif ac_type == "cargo":  # all cargo
        W_instruments = 800.0 * Units.lb
        W_avionics = 900.0 * Units.lb
        W_electrical = 1950.0 * Units.lb  # for cargo a/c
    elif ac_type == "commuter":  # commuter
        W_instruments = 300.0 * Units.lb
        W_avionics = 500.0 * Units.lb
    elif ac_type == "sst":  # sst
        W_instruments = 1200.0 * Units.lb
        W_avionics = 1500.0 * Units.lb
        W_furnish += 23.0 * num_seats * Units.lb  # add aditional seat wt
    else:
        W_instruments = 800.0 * Units.lb
        W_avionics = 900.0 * Units.lb 

    # packup outputs
    output = Data()
    output.W_flight_control    = W_flight_controls
    output.W_apu               = apu_wt
    output.W_hyd_pnu           = hyd_pnu_wt
    output.W_instruments       = W_instruments
    output.W_avionics          = W_avionics
    output.W_electrical        = W_electrical
    output.W_ac                = W_air_conditioning
    output.W_furnish           = W_furnish
    output.W_anti_ice          = 0 # included in AC
    output.W_systems           = output.W_flight_control + output.W_apu + output.W_hyd_pnu \
                                + output.W_ac + output.W_avionics + output.W_electrical \
                                + output.W_furnish + output.W_instruments

    return output



# RCAIDE/Methods/Noise/Correlation_Buildup/Engine/jet_installation_effect.py
# 
# 
# Created:  Jul 2023, M. Clarke  

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
# Python package imports   
import numpy as np   

# ----------------------------------------------------------------------------------------------------------------------     
#  Jet Installation Effect
# ----------------------------------------------------------------------------------------------------------------------      
def jet_installation_effect(Xe,Ye,Ce,theta_s,Diameter_mixed):
    """This calculates the installation effect, in decibels, to be added to the predicted secondary jet noise level.
    
    Assumptions:
        N/A

    Source:
        SAE ARP876D: Gas Turbine Jet Exhaust Noise Prediction

    Inputs:
        Ce = wing chord length at the engine location - as figure 7.3 of the SAE ARP 876D                    [m]                          
        Xe = fan exit location downstream of the leading edge (Xe<Ce) - as figure 7.3 of the SAE ARP 876D    [m] 
        Ye = separation distance from the wing chord line to nozzle lip - as figure 7.3 of the SAE ARP 876D  [m] 
        theta_s                                                                                              [rad]
        Diameter_mixed                                                                                       [m] 

    Outputs:
        INST_s          [-]

    Properties Used: 
        N/A 
    """ 
    # Instalation effect
    INST_s = 0.5*((Ce-Xe)**2/(Ce*Diameter_mixed))*(np.exp(-Ye/Diameter_mixed)*((1.8*theta_s/np.pi))-0.6)**2  
    INST_s[INST_s>2.5]=2.5 

    return INST_s


# RCAIDE/Library/Plots/Performance/Aerodynamics/plot_aerodynamic_forces.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np 

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------   
def plot_aerodynamic_forces(results,
                             save_figure = False,
                             show_legend = True,
                             save_filename = "Aerodynamic_Forces",
                             file_type = ".png",
                             width = 11, height = 7):
    """This plots the aerodynamic forces
    
    Assumptions:
    None
    
    Source:
    None
    
    Inputs:
    results.segments.condtions.frames
         body.thrust_force_vector
         wind.force_vector
         wind.force_vector
         
    Outputs:
    Plots
    
    Properties Used:
    N/A
    """ 

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters)
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))     
    
    fig   = plt.figure(save_filename)
    fig.set_size_inches(width,height)
    
    for i in range(len(results.segments)): 
        time   = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min
        Power  = results.segments[i].conditions.energy.power[:,0] 
        Thrust = results.segments[i].conditions.frames.body.thrust_force_vector[:,0]
        Lift   = -results.segments[i].conditions.frames.wind.force_vector[:,2]
        Drag   = -results.segments[i].conditions.frames.wind.force_vector[:,0]
         
        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')
        
        # power 
        axis_1 = plt.subplot(2,2,1)
        axis_1.set_ylabel(r'Power (MW)')
        axis_1.plot(time,Power/1E6, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width, label = segment_name) 
        set_axes(axis_1)                
        
        axis_2 = plt.subplot(2,2,2)
        axis_2.plot(time, Thrust/1000, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width) 
        axis_2.set_ylabel(r'Thrust (kN)')
        set_axes(axis_2) 

        axis_3 = plt.subplot(2,2,3)
        axis_3.plot(time, Lift/1000, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_3.set_xlabel('Time (mins)')
        axis_3.set_ylabel(r'Lift (kN)')
        set_axes(axis_3) 
        
        axis_4 = plt.subplot(2,2,4)
        axis_4.plot(time,Drag/1000 , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width)
        axis_4.set_xlabel('Time (mins)')
        axis_4.set_ylabel(r'Drag (kN)')
        set_axes(axis_4)  
        
 
    if show_legend:    
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})    
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Aerodynamic Forces'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
    return fig 




""" RCAIDE Package Setup for rotor wake methods that are directly specified by analyses.
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Wake.Perscribed_Vortex_Wake.compute_fidelity_zero_induced_velocity   import compute_fidelity_zero_induced_velocity 
from RCAIDE.Library.Methods.Propulsors.Converters.Rotor.Wake.Perscribed_Vortex_Wake.compute_wake_contraction_matrix          import compute_wake_contraction_matrix


# RCAIDE/Compoments/Wings/Control_Surfaces/Rudder.py
# 
# Created:  Mar 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
# RCAIDE imports   
from .Control_Surface import Control_Surface 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Rudder
# ----------------------------------------------------------------------------------------------------------------------
class Rudder(Control_Surface):
    """
    A class representing a rudder control surface for aircraft directional control.

    Attributes
    ----------
    tag : str
        Unique identifier for the rudder, defaults to 'rudder'
        
    hinge_fraction : float
        Location of the hinge line as fraction of chord, defaults to 0.0
        
    sign_duplicate : float
        Sign convention for duplicate rudder deflection, defaults to 1.0
        (typically only one rudder per aircraft)

    Notes
    -----
    The rudder is a trailing edge control surface mounted on the vertical tail, used 
    primarily for yaw control. It inherits basic control surface functionality from 
    the Control_Surface class and adds specific attributes for rudder operation.
    
    **Definitions**

    'Hinge Fraction'
        The chordwise location of the rudder hinge line, measured from the leading 
        edge as a fraction of local chord
        
    'Sign Duplicate'
        Determines whether paired rudders (if present) deflect in the same or 
        opposite directions. Typically 1.0 as most aircraft have a single rudder

    See Also
    --------
    RCAIDE.Library.Components.Wings.Control_Surfaces.Control_Surface
        Base class providing common control surface functionality
    RCAIDE.Library.Components.Wings.Vertical_Tail_All_Moving
        Alternative yaw control using all-moving vertical tail
    """ 

    def __defaults__(self):
        """
        Sets default values for the rudder attributes.
        
        Notes
        -----
        See Control_Surface.__defaults__ for additional inherited attributes.
        """
        self.tag            = 'rudder'
        self.hinge_fraction = 0.0
        self.sign_duplicate = 1.0
        
        pass 


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/General_Aviation/compute_fuselage_weight.py
# 
# 
# Created:  Sep 2024, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

# RCAIDE 
from RCAIDE.Framework.Core import  Units
import  numpy as  np

# ----------------------------------------------------------------------------------------------------------------------
# Fuselage Weight 
# ----------------------------------------------------------------------------------------------------------------------
def compute_fuselage_weight(S_fus, Nult, TOW, w_fus, h_fus, l_fus, l_ht, q_c, V_fuse, diff_p_fus):
    """ 
        Calculate the weight of a fuselage for a GA aircraft
        
        Source: Raymer: Aircraft Design, a Conceptual Approach (pages 460-461 in 4th edition)
        
        Inputs:
            S_f - fuselage wetted area                          [meters**2]
            Nult - ultimate load of the aircraft                [dimensionless]]
            TOW - maximum takeoff weight of the aircraft        [kilograms]
            w_fus - width of the fuselage                       [meters]
            h_fus - height of the fuselage                      [meters]
            l_fus - length of the fuselage                      [meters]
            l_ht  - length of tail arm                          [meters]
            q_c    - dynamic pressure at cruise                 [Pascals] 
            V_fuse - volume of pressurized cabin                [meters**3]
            diff_p_fus - Maximum fuselage pressure differential [Pascals]
           

            
        Outputs:
            weight - weight of the fuselage [kilograms]
            
         Assumptions:
            fuselage for a general aviation type aircraft
    """ 
    # take average as diameter
    d_fus    = (h_fus+w_fus)/2.
    
    # obtained from http://homepage.ntlworld.com/marc.barbour/structures.html
    d_str    = .025*d_fus+1.*Units.inches
    
    diff_p   = diff_p_fus / (Units.force_pound / Units.ft**2.) # Convert Pascals to lbs/ square ft 
    tail_arm = np.abs(l_ht)/Units.ft 
    weight   = TOW / Units.lb    # Convert kg to lbs
    area     = S_fus / (Units.ft**2.) # Convert square meters to square ft 
    q        = q_c /(Units.force_pound / Units.ft**2.)

    # Calculate weight of wing for traditional aircraft vertical tail without rudder
    fuselage_weight = .052*(area**1.086)*((Nult*weight)**.177)*(tail_arm**(-.051))*((l_fus/d_str)**(-.072))*(q**.241)+11.9*((V_fuse* diff_p)**.271)

    weight = fuselage_weight*Units.lbs #convert to kg
    return weight


# RCAIDE/Library/Plots/Performance/Mission/plot_flight_trajectory.py
# 
# 
# Created:  Jul 2023, M. Clarke 

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  
from RCAIDE.Framework.Core import Units
from RCAIDE.Library.Plots.Common import set_axes, plot_style
import matplotlib.pyplot as plt
import matplotlib.cm as cm
import numpy as np  

# ----------------------------------------------------------------------------------------------------------------------
#  PLOTS
# ----------------------------------------------------------------------------------------------------------------------              
def plot_flight_trajectory(results,
                           line_color = 'bo-',
                           line_color2 = 'rs--',
                           save_figure = False,
                           show_legend   = True,
                           save_filename = "Flight_Trajectory",
                           file_type = ".png",
                           width = 11, height = 7):
    """This plots the 3D flight trajectory of the aircraft.

    Assumptions:
    None

    Source:
    None

    Inputs:
    results.segments.conditions.
         frames 
             body.inertial_rotations
             inertial.position_vector 
         freestream.velocity
         aerodynamics.
             lift_coefficient
             drag_coefficient
             angle_of_attack
        
    Outputs: 
    Plots

    Properties Used:
    N/A	
    """	     

    # get plotting style 
    ps      = plot_style()  

    parameters = {'axes.labelsize': ps.axis_font_size,
                  'xtick.labelsize': ps.axis_font_size,
                  'ytick.labelsize': ps.axis_font_size,
                  'axes.titlesize': ps.title_font_size}
    plt.rcParams.update(parameters) 
        
    fig = plt.figure(save_filename)
    fig.set_size_inches(width,height) 
     
    # get line colors for plots 
    line_colors   = cm.inferno(np.linspace(0,0.9,len(results.segments)))    
     
    for i in range(len(results.segments)): 
        time     = results.segments[i].conditions.frames.inertial.time[:,0] / Units.min
        Range    = results.segments[i].conditions.frames.inertial.aircraft_range[:,0]/Units.nmi
        x        = results.segments[i].conditions.frames.inertial.position_vector[:,0]  
        y        = results.segments[i].conditions.frames.inertial.position_vector[:,1] 
        z        = -results.segments[i].conditions.frames.inertial.position_vector[:,2] 

        segment_tag  =  results.segments[i].tag
        segment_name = segment_tag.replace('_', ' ')
        
        axes = plt.subplot(2,2,1)
        axes.plot( time , Range, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width , label = segment_name)
        axes.set_ylabel('Distance (nmi)')
        axes.set_xlabel('Time (min)')
        set_axes(axes)            

        axes = plt.subplot(2,2,2)
        axes.plot(x, y , line_color, color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width )
        axes.set_xlabel('x (m)')
        axes.set_ylabel('y (m)')
        set_axes(axes)

        axes = plt.subplot(2,2,3)
        axes.plot( time , z, line_color , color = line_colors[i], marker = ps.markers[0], linewidth = ps.line_width )
        axes.set_ylabel('z (m)')
        axes.set_xlabel('Time (min)')
        set_axes(axes)   
        
        axes = plt.subplot(2,2,4, projection='3d') 
        axes.scatter(x, y, z, marker='o',color =  line_colors[i])
        axes.set_xlabel('x')
        axes.set_ylabel('y')
        axes.set_zlabel('z')
        axes.set_box_aspect([1,1,1])
        set_axes(axes)         
        
    if show_legend:        
        leg =  fig.legend(bbox_to_anchor=(0.5, 0.95), loc='upper center', ncol = 4) 
        leg.set_title('Flight Segment', prop={'size': ps.legend_font_size, 'weight': 'heavy'})    
    
    # Adjusting the sub-plots for legend 
    fig.tight_layout()
    fig.subplots_adjust(top=0.8)
    
    # set title of plot 
    title_text    = 'Flight Trajectory'      
    fig.suptitle(title_text)
    
    if save_figure:
        plt.savefig(save_filename + file_type)   
             
    return fig         


# RCAIDE/Methods/Energy/Propulsors/Turbofan_Propulsor/compute_thrust.py
# 
# 
# Created:  Jul 2023, M. Clarke

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ---------------------------------------------------------------------------------------------------------------------- 
 # RCAIDE imports  
from RCAIDE.Framework.Core      import Units 

# Python package imports
import numpy as np
 
# ----------------------------------------------------------------------------------------------------------------------
#  compute_thrust
# ----------------------------------------------------------------------------------------------------------------------
def compute_thrust(turbofan,turbofan_conditions,conditions):
    """Computes thrust and other properties of the turbofan listed below: 
    turbofan.  
      .outputs.thrust                                    (numpy.ndarray): thrust                               [N] 
      .outputs.thrust_specific_fuel_consumption          (numpy.ndarray): TSFC                                 [N/N-s] 
      .outputs.non_dimensional_thrust                    (numpy.ndarray): non-dim thurst                       [unitless] 
      .outputs.core_mass_flow_rate                       (numpy.ndarray): core nozzle mass flow rate           [kg/s] 
      .outputs.fuel_flow_rate                            (numpy.ndarray): fuel flow rate                       [kg/s] 
      .outputs.power                                     (numpy.ndarray): power                                [W] 
      
    Assumptions:
        Perfect gas

    Source:
        Stanford AA 283 Course Notes: https://web.stanford.edu/~cantwell/AA283_Course_Material/AA283_Course_Notes/


    Args: 
        conditions. 
           freestream.isentropic_expansion_factor                (float): isentropic expansion factor          [unitless]  
           freestream.specific_heat_at_constant_pressure         (float): speific heat                         [J/(kg K)] 
           freestream.velocity                           (numpy.ndarray): freestream velocity                  [m/s] 
           freestream.speed_of_sound                     (numpy.ndarray): freestream speed_of_sound            [m/s] 
           freestream.mach_number                        (numpy.ndarray): freestream mach_number               [unitless] 
           freestream.pressure                           (numpy.ndarray): freestream pressure                  [Pa] 
           freestream.gravity                            (numpy.ndarray): freestream gravity                   [m/s^2] 
           propulsion.throttle                           (numpy.ndarray): throttle                             [unitless] 
        turbofan 
           .fuel_to_air_ratio                                    (float): fuel_to_air_ratio                    [unitless] 
           .total_temperature_reference                          (float): total_temperature_reference          [K] 
           .total_pressure_reference                             (float): total_pressure_reference             [Pa]    
           .core_nozzle.velocity                         (numpy.ndarray): turbofan core nozzle velocity        [m/s] 
           .core_nozzle.static_pressure                  (numpy.ndarray): turbofan core nozzle static pressure [Pa] 
           .core_nozzle.area_ratio                               (float): turbofan core nozzle area ratio      [unitless] 
           .fan_nozzle.velocity                          (numpy.ndarray): turbofan fan nozzle velocity         [m/s] 
           .fan_nozzle.static_pressure                   (numpy.ndarray): turbofan fan nozzle static pressure  [Pa] 
           .fan_nozzle.area_ratio                                (float): turbofan fan nozzle area ratio       [unitless]   
           .reference_temperature                                (float): reference_temperature                [K] 
           .reference_pressure                                   (float): reference_pressure                   [Pa] 
           .compressor_nondimensional_massflow                   (float): non-dim mass flow rate               [unitless]
      
    Returns:
        None
         
    """      
    # Unpack flight conditions 
    gamma                       = conditions.freestream.isentropic_expansion_factor 
    u0                          = conditions.freestream.velocity
    a0                          = conditions.freestream.speed_of_sound
    M0                          = conditions.freestream.mach_number
    p0                          = conditions.freestream.pressure  
    g                           = conditions.freestream.gravity        

    # Unpack turbofan operating conditions and properties 
    Tref                        = turbofan.reference_temperature
    Pref                        = turbofan.reference_pressure
    mdhc                        = turbofan.compressor_nondimensional_massflow
    SFC_adjustment              = turbofan.SFC_adjustment 
    f                           = turbofan_conditions.fuel_to_air_ratio
    total_temperature_reference = turbofan_conditions.total_temperature_reference
    total_pressure_reference    = turbofan_conditions.total_pressure_reference 
    flow_through_core           = turbofan_conditions.flow_through_core 
    flow_through_fan            = turbofan_conditions.flow_through_fan  
    V_fan_nozzle                = turbofan_conditions.fan_nozzle_exit_velocity
    fan_area_ratio              = turbofan_conditions.fan_nozzle_area_ratio
    P_fan_nozzle                = turbofan_conditions.fan_nozzle_static_pressure
    P_core_nozzle               = turbofan_conditions.core_nozzle_static_pressure
    V_core_nozzle               = turbofan_conditions.core_nozzle_exit_velocity
    core_area_ratio             = turbofan_conditions.core_nozzle_area_ratio                   
    bypass_ratio                = turbofan_conditions.bypass_ratio  

    # Compute  non dimensional thrust
    fan_thrust_nondim   = flow_through_fan*(gamma*M0*M0*(V_fan_nozzle/u0-1.) + fan_area_ratio*(P_fan_nozzle/p0-1.))
    core_thrust_nondim  = flow_through_core*(gamma*M0*M0*(V_core_nozzle/u0-1.) + core_area_ratio*(P_core_nozzle/p0-1.))

    thrust_nondim       = core_thrust_nondim + fan_thrust_nondim

    # Computing Specifc Thrust
    Fsp   = 1./(gamma*M0)*thrust_nondim
    Fsp_c = 1./(gamma*M0)*core_thrust_nondim
    Fsp_f = 1./(gamma*M0)*fan_thrust_nondim

    # Compute specific impulse
    Isp   = Fsp*a0*(1.+bypass_ratio)/(f*g)

    # Compute TSFC
    TSFC  = f*g/(Fsp*a0*(1.+bypass_ratio))*(1.-SFC_adjustment) * Units.hour # 1/s is converted to 1/hr here
 
    # Compute core mass flow
    mdot_core  = mdhc*np.sqrt(Tref/total_temperature_reference)*(total_pressure_reference/Pref)

    # Compute dimensional thrust
    FD2   = Fsp*a0*(1.+bypass_ratio)*mdot_core*turbofan_conditions.throttle
    FD2_f = Fsp_f*a0*(1.+bypass_ratio)*mdot_core*turbofan_conditions.throttle
    FD2_c = Fsp_c*a0*(1.+bypass_ratio)*mdot_core*turbofan_conditions.throttle

    # Compute power 
    power   = FD2*u0    

    # Compute fuel flow rate 
    fuel_flow_rate   = np.fmax(FD2*TSFC/g,np.array([0.]))*1./Units.hour

    # Pack turbofan outouts  
    turbofan_conditions.thrust                            = FD2 
    turbofan_conditions.fan_thrust                        = FD2_f 
    turbofan_conditions.core_thrust                       = FD2_c 
    turbofan_conditions.thrust_specific_fuel_consumption  = TSFC
    turbofan_conditions.non_dimensional_thrust            = Fsp  
    turbofan_conditions.power                             = power  
    turbofan_conditions.specific_impulse                  = Isp
    turbofan_conditions.core_mass_flow_rate               = mdot_core
    turbofan_conditions.fuel_flow_rate                    = fuel_flow_rate   
    
    return  


# RCAIDE/Library/Components/Propulsors/__init__.py
# 
# 
"""
Collection of propulsion system components for aircraft modeling and simulation.

This module provides various propulsor classes for different types of aircraft 
propulsion systems, including:

- Electric propulsion (Electric_Rotor, Electric_Ducted_Fan)
- Internal combustion engine propulsion (ICE_Propeller, Constant_Speed_ICE_Propeller)
- Gas turbine propulsion (Turbofan, Turbojet, Turboshaft, Turboprop)
- Power conversion components (Converters)

Notes
-----
Each propulsor class inherits from the base Propulsor class and implements 
specific methods for performance calculation, operating condition management, 
and system state tracking.

The module structure allows for easy extension with new propulsion system types
while maintaining consistent interfaces across all implementations.

See Also
--------
RCAIDE.Library.Components.Energy
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .                                           import Converters  
from .Propulsor                                  import Propulsor
from .Electric_Rotor                             import Electric_Rotor
from .Electric_Ducted_Fan                        import Electric_Ducted_Fan
from .ICE_Propeller                              import ICE_Propeller
from .Turbofan                                   import Turbofan
from .Turbojet                                   import Turbojet
from .Turboshaft                                 import Turboshaft
from .Turboprop                                  import Turboprop
from .Constant_Speed_ICE_Propeller               import Constant_Speed_ICE_Propeller



## @ingroup Analyses-Atmospheric
# US_Standard_1976.py
#
# Created: 
# Modified: Feb 2016, Andrew Wendorff
#           Jan 2018, W. Maier

# ----------------------------------------------------------------------
#  Imports
# ----------------------------------------------------------------------

import numpy as np
from warnings import warn
import RCAIDE
from RCAIDE.Framework.Analyses.Atmospheric import Atmospheric
from RCAIDE.Framework.Mission.Common.Conditions import Conditions
from RCAIDE.Framework.Core import Units
from RCAIDE.Framework.Core.Arrays import atleast_2d_col

from RCAIDE.Library.Attributes.Gases import Air
from RCAIDE.Library.Attributes.Planets import Earth
 

# ----------------------------------------------------------------------
#  Classes
# ----------------------------------------------------------------------

## @ingroup Analyses-Atmospheric
class US_Standard_1976(Atmospheric):

    """ Implements the U.S. Standard Atmosphere (1976 version)
        
    Assumptions:
    None
    
    Source:
    U.S. Standard Atmosphere, 1976, U.S. Government Printing Office, Washington, D.C., 1976
    """
    
    def __defaults__(self):
        """This sets the default values for the analysis to function.

        Assumptions:
        None

        Source:
        N/A

        Inputs:
        None

        Output:
        None

        Properties Used:
        None
        """     
        
        atmo_data = RCAIDE.Library.Attributes.Atmospheres.Earth.US_Standard_1976()
        self.update(atmo_data)        
    
    def compute_values(self,altitude,temperature_deviation=0.0,var_gamma=False):

        """Computes atmospheric values.

        Assumptions:
        US 1976 Standard Atmosphere

        Source:
        U.S. Standard Atmosphere, 1976, U.S. Government Printing Office, Washington, D.C., 1976

        Inputs:
        altitude                                 [m]
        temperature_deviation                    [K]

        Output:
        atmo_data.
          pressure                               [Pa]
          temperature                            [K]
          speed_of_sound                         [m/s]
          dynamic_viscosity                      [kg/(m*s)]
          kinematic_viscosity                    [m^2/s]
          thermal_conductivity                   [W/(m*K)]
          prandtl_number                         [-]
           
        Properties Used:
        self.
          fluid_properties.gas_specific_constant [J/(kg*K)]
          planet.sea_level_gravity               [m/s^2]
          planet.mean_radius                     [m]
          breaks.
            altitude                             [m]
            temperature                          [K]
            pressure                             [Pa]
        """

        # unpack
        zs        = altitude
        gas       = self.fluid_properties
        planet    = self.planet
        grav      = self.planet.sea_level_gravity        
        Rad       = self.planet.mean_radius
        R         = gas.gas_specific_constant
        delta_isa = temperature_deviation
        
        # check properties
        if not gas == Air():
            warn('US Standard Atmosphere not using Air fluid properties')
        if not planet == Earth():
            warn('US Standard Atmosphere not using Earth planet properties')          
        
        # convert input if necessary
        zs = atleast_2d_col(zs)

        # get model altitude bounds
        zmin = self.breaks.altitude[0]
        zmax = self.breaks.altitude[-1]   
        
        # convert geometric to geopotential altitude
        zs = zs/(1 + zs/Rad)
        
        # check ranges
        if np.amin(zs) < zmin:
            print("Warning: altitude requested below minimum for this atmospheric model; returning values for h = -2.0 km")
            zs[zs < zmin] = zmin
        if np.amax(zs) > zmax:
            print("Warning: altitude requested above maximum for this atmospheric model; returning values for h = 86.0 km")   
            zs[zs > zmax] = zmax        

        # initialize return data
        zeros = np.zeros_like(zs)
        p     = zeros * 0.0
        T     = zeros * 0.0
        rho   = zeros * 0.0
        a     = zeros * 0.0
        mu    = zeros * 0.0
        z0    = zeros * 0.0
        T0    = zeros * 0.0
        p0    = zeros * 0.0
        alpha = zeros * 0.0
        
        # populate the altitude breaks
        # this uses >= and <= to capture both edges and because values should be the same at the edges
        for i in range( len(self.breaks.altitude)-1 ): 
            i_inside = (zs >= self.breaks.altitude[i]) & (zs <= self.breaks.altitude[i+1])
            z0[ i_inside ]    = self.breaks.altitude[i]
            T0[ i_inside ]    = self.breaks.temperature[i]
            p0[ i_inside ]    = self.breaks.pressure[i]
            alpha[ i_inside ] = -(self.breaks.temperature[i+1] - self.breaks.temperature[i])/ \
                                 (self.breaks.altitude[i+1]    - self.breaks.altitude[i])
        
        # interpolate the breaks
        dz = zs-z0
        i_isoth = (alpha == 0.)
        i_adiab = (alpha != 0.)
        p[i_isoth] = p0[i_isoth] * np.exp(-1.*dz[i_isoth]*grav/(R*T0[i_isoth]))
        p[i_adiab] = p0[i_adiab] * ( (1.-alpha[i_adiab]*dz[i_adiab]/T0[i_adiab]) **(1.*grav/(alpha[i_adiab]*R)) )
        
        T     = T0 - dz*alpha + delta_isa
        rho   = gas.compute_density(T,p)
        a     = gas.compute_speed_of_sound(T,p,var_gamma)
        mu    = gas.compute_absolute_viscosity(T)
        K     = gas.compute_thermal_conductivity(T)  
        Pr    = gas.compute_prandtl_number(T)
        
        atmo_data = Conditions()
        atmo_data.expand_rows(zs.shape[0])
        atmo_data.pressure                     = p
        atmo_data.temperature                  = T
        atmo_data.density                      = rho
        atmo_data.speed_of_sound               = a
        atmo_data.dynamic_viscosity            = mu
        atmo_data.kinematic_viscosity          = mu/rho
        atmo_data.thermal_conductivity         = K
        atmo_data.prandtl_number               = Pr 
        
        return atmo_data


# ----------------------------------------------------------------------
#   Module Tests
# ----------------------------------------------------------------------
if __name__ == '__main__':
    
    import pylab as plt
    
    h = np.linspace(-1.,60.,200) * Units.km
    delta_isa = 0.
    h = 5000.
    atmosphere = US_Standard_1976()
    
    data = atmosphere.compute_values(h,delta_isa)
    p   = data.pressure
    T   = data.temperature
    rho = data.density
    a   = data.speed_of_sound
    mu  = data.dynamic_viscosity
    K   = data.thermal_conductivity
    
    print(data)
    


# RCAIDE/Library/Methods/Weights/Correlations/Propulsion/__init__.py
# 

"""RCAIDE Package Setup
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------
 
from .compute_jet_engine_weight                             import compute_jet_engine_weight
from .compute_piston_engine_weight                          import compute_piston_engine_weight 
from .integrated_propulsion                                 import integrated_propulsion
from .integrated_propulsion_general_aviation                import integrated_propulsion_general_aviation
from .compute_motor_weight                                  import compute_motor_weight
from RCAIDE.Library.Methods.Weights.Physics_Based_Buildups.Electric   import dynamo_supply_mass_estimation


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Common/compute_landing_gear_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Data 

# ---------------------------------------------------------------------------------------------------------------------- 
# Landing Gear 
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_landing_gear_weight(vehicle,landing_gear_W_factor=0.04):
    """ Calculate the weight of the landing gear assuming that the gear 
    weight is 4 percent of the takeoff weight        
    
    Assumptions:
        calculating the landing gear weight based on the takeoff weight
    
    Source: 
        N/A
        
    Inputs:
        TOW - takeoff weight of the aircraft                              [kilograms]
        landing_gear_W_factor - landing gear weight as percentage of TOW [dimensionless]
    
    Outputs:
        weight - weight of the landing gear                               [kilograms]
            
    Properties Used:
        N/A
    """

    # process
    weight          = landing_gear_W_factor * vehicle.mass_properties.max_takeoff
    output          = Data()
    output.main     = weight * 0.9
    output.nose     = weight * 0.1
    return output



# RCAIDE/Library/Components/Wings/__init__.py

"""
Module containing wing and lifting surface components for aircraft design and analysis. 
This module provides classes for various wing types including main wings, vertical and 
horizontal tails, and all-moving surfaces.

The module also includes a Control_Surfaces subpackage for various control surface types.

See Also
--------
RCAIDE.Library.Components.Airfoils
    Related module for airfoil definitions used in wing design
RCAIDE.Library.Components.Fuselages
    Related module for fuselage components that interface with wings
"""

# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------

from .Wing                      import Wing
from .Main_Wing                 import Main_Wing
from .Vertical_Tail             import Vertical_Tail
from .Horizontal_Tail           import Horizontal_Tail
from .All_Moving_Surface        import All_Moving_Surface
from .Stabilator                import Stabilator
from .Vertical_Tail_All_Moving  import Vertical_Tail_All_Moving

# packages
from . import Control_Surfaces
from . import Segments


# RCAIDE/Methods/Energy/Propulsors/Rotor_Design/optimization_setup.py
# 
# 
# Created:  Jul 2023, M. Clarke 


# ----------------------------------------------------------------------------------------------------------------------
#  IMPORT
# ----------------------------------------------------------------------------------------------------------------------  

# RCAIDE Imports 
import RCAIDE   
from RCAIDE.Library.Methods.Geometry.Airfoil  import compute_airfoil_properties, compute_naca_4series, import_airfoil_geometry

# Python package imports   
import numpy as np  
    
# ----------------------------------------------------------------------------------------------------------------------  
#  Blade Geometry Setup 
# ----------------------------------------------------------------------------------------------------------------------    
def blade_geometry_setup(rotor,number_of_stations): 
    """ Defines a dummy vehicle for prop-rotor blade optimization.
          
          Inputs:  
             rotor   - rotor data structure             [None] 
              
          Outputs:  
             configs - configuration used in optimization    [None]
              
          Assumptions: 
             N/A 
        
          Source:
             None
    """    
    
    # Unpack prop-rotor geometry  
    N                     = number_of_stations       
    B                     = rotor.number_of_blades    
    R                     = rotor.tip_radius
    Rh                    = rotor.hub_radius 
    design_thrust_hover   = rotor.hover.design_thrust
    design_power_hover    = rotor.hover.design_power
    chi0                  = Rh/R  
    chi                   = np.linspace(chi0,1,N+1)  
    chi                   = chi[0:N]
    airfoils              = rotor.airfoils      
    a_loc                 = rotor.airfoil_polar_stations  
    
    # determine target values 
    if (design_thrust_hover == None) and (design_power_hover== None):
        raise AssertionError('Specify either design thrust or design power at hover!') 
    elif (design_thrust_hover!= None) and (design_power_hover!= None):
        raise AssertionError('Specify either design thrust or design power at hover!')     
    num_airfoils = len(airfoils.keys())
    if num_airfoils>0:
        if len(a_loc) != N:
            raise AssertionError('\nDimension of airfoil sections must be equal to number of stations on propeller') 
        
        for _,airfoil in enumerate(airfoils):  
            if airfoil.geometry == None: # first, if airfoil geometry data not defined, import from geoemtry files
                if type(airfoil) == RCAIDE.Library.Components.Airfoils.NACA_4_Series_Airfoil: # check if naca 4 series of airfoil from datafile
                    airfoil.geometry = compute_naca_4series(airfoil.NACA_4_Series_code,airfoil.number_of_points)
                else:
                    airfoil.geometry = import_airfoil_geometry(airfoil.coordinate_file,airfoil.number_of_points) 
    
            if airfoil.polars == None: # compute airfoil polars for airfoils
                airfoil.polars = compute_airfoil_properties(airfoil.geometry, airfoil_polar_files= airfoil.polar_files) 
                     
    # thickness to chord         
    t_c           = np.zeros(N)    
    if num_airfoils>0:
        for j,airfoil in enumerate(airfoils): 
            a_geo         = airfoil.geometry
            locs          = np.where(np.array(a_loc) == j ) 
            t_c[locs]     = a_geo.thickness_to_chord 
            
    # append additional prop-rotor  properties for optimization  
    rotor.number_of_blades             = int(B)  
    rotor.thickness_to_chord           = t_c
    rotor.radius_distribution          = chi*R  
    
    # set oei conditions if they are not set
    if rotor.oei.design_freestream_velocity == None: 
        rotor.oei.design_freestream_velocity = rotor.hover.design_freestream_velocity
    if rotor.oei.design_altitude == None:
        rotor.oei.design_altitude = rotor.hover.design_altitude
    if design_thrust_hover == None:
        if rotor.oei.design_power == None: 
            rotor.oei.design_power = rotor.hover.design_power*1.1
    elif  design_power_hover == None:
        if rotor.oei.design_thrust == None: 
            rotor.oei.design_thrust = rotor.hover.design_thrust*1.1
    
    vehicle                            = RCAIDE.Vehicle()  
    net                                = RCAIDE.Framework.Networks.Electric() 
    bus                                = RCAIDE.Library.Components.Energy.Distributors.Electrical_Bus()
    propulsor                          = RCAIDE.Library.Components.Propulsors.Electric_Rotor()
    propulsor.rotor                    = rotor
    bus.assigned_propulsors = [[propulsor.tag]]
    net.propulsors.append(propulsor)
    net.busses.append(bus)
    vehicle.append_energy_network(net)
    
    configs                             = RCAIDE.Library.Components.Configs.Config.Container()
    base_config                         = RCAIDE.Library.Components.Configs.Config(vehicle) 
    
    config                              = RCAIDE.Library.Components.Configs.Config(base_config)
    config.tag                          = 'hover' 
    config.networks.electric.propulsors.electric_rotor.rotor.orientation_euler_angles = [0.0,np.pi/2,0.0]    
    configs.append(config)        

    config                              = RCAIDE.Library.Components.Configs.Config(base_config)
    config.tag                          = 'oei' 
    config.networks.electric.propulsors.electric_rotor.rotor.orientation_euler_angles = [0.0,np.pi/2,0.0]    
    configs.append(config)       
    
    if type(rotor) == RCAIDE.Library.Components.Propulsors.Converters.Prop_Rotor:  
        design_thrust_cruise  = rotor.cruise.design_thrust 
        design_power_cruise   = rotor.cruise.design_power      
        if (design_thrust_cruise == None) and (design_power_cruise== None):
            raise AssertionError('Specify either design thrust or design power at cruise!') 
        elif (design_thrust_cruise!= None) and (design_power_cruise!= None):
            raise AssertionError('Specify either design thrust or design power at cruise!') 
        
        config                          = RCAIDE.Library.Components.Configs.Config(base_config)
        config.tag                      = 'cruise'
        config.networks.electric.propulsors.electric_rotor.rotor.orientation_euler_angles = [0.0,np.pi/2,0.0] 
        configs.append(config)
    return configs 


## @ingroup Methods-Propulsion
# nozzle_calculations.py
# 
# Created:  Sep 2017, P. Goncalves
# Modified: 

import numpy as np
from scipy.optimize import fsolve

# ----------------------------------------------------------------------
#  nozzle calculations
# ----------------------------------------------------------------------

## @ingroup Methods-Propulsion
def exit_Mach_shock(area_ratio, gamma, Pt_out, P0):
    """ Determines the output Mach number of a nozzle with a normal shock taking
    place inside of it, through pressure ratio between the nozzle stagnation
    pressure and the freestream pressure
    
    Assumptions:
    Unknown
    
    Source:
    Unknown

    Inputs:
    area_ratio    [dimensionless]
    gamma         [dimensionless]
    Pt_out        [Pascals]
    P0            [Pascals]
    
    Outputs:
    Me            [dimensionless]      
    
    """
    func = lambda Me : (Pt_out/P0)*(1./area_ratio)-(((gamma+1.)/2.)**((gamma+1.)/(2.*(gamma-1.))))*Me*((1.+(gamma-1.)/2.*Me**2.)**0.5)

    #Initializing the array
    Me_initial_guess = np.ones_like(Pt_out) 
    i_sol = Me_initial_guess < 10.0
    Me_initial_guess[i_sol] = 0.1

    # Solving for Me
    Me = fsolve(func,Me_initial_guess)
        
    return Me
        
## @ingroup Methods-Propulsion
def mach_area(area_ratio, gamma, subsonic):
    """ Returns the Mach number given an area ratio and isentropic conditions
    
    Assumptions:
    Unknown
    
    Source:
    Unknown

    Inputs:
    area_ratio    [dimensionless]
    gamma         [dimensionless]
    subsonic      [Boolean]
    
    Outputs:
    Me            [dimensionless]  
    
    """
    func = lambda Me : (area_ratio**2. - ((1./Me)**2.)*(((2./(gamma+1.))*(1.+((gamma-1.)/2.)*Me**2.))**((gamma+1.)/((gamma-1.)))))[:,0]
    if subsonic:
        Me_initial_guess = np.ones_like(gamma)*0.01
    else:
        Me_initial_guess = np.ones_like(gamma)*2.0         
        
    Me = np.atleast_2d(fsolve(func,Me_initial_guess, factor = 0.1)).T

    return Me

    
## @ingroup Methods-Propulsion
def normal_shock(M1, gamma):  
    """ Returns the Mach number after normal shock
    
    Assumptions:
    Unknown
    
    Source:
    Unknown

    Inputs:
    M1            [dimensionless]
    gamma         [dimensionless]
    
    Outputs:
    M2            [dimensionless]      
    
    """
    M2 = np.sqrt((((gamma-1.)*M1**2.)+2.)/(2.*gamma*M1**2.-(gamma-1.)))
    
    return M2

## @ingroup Methods-Propulsion
def pressure_ratio_isentropic(area_ratio, gamma, subsonic):
    """ Determines the pressure ratio for isentropic flow throughout the entire
    nozzle
    
    Assumptions:
    Unknown
    
    Source:
    Unknown

    Inputs:
    area_ratio    [dimensionless]
    gamma         [dimensionless]
    subsonic      [Boolean]
    
    Outputs:
    pr_isentropic [dimensionless]      
        
    """
    #yields pressure ratio for isentropic conditions given area ratio
    Me = mach_area(area_ratio,gamma, subsonic)
    
    pr_isentropic = (1.+((gamma-1.)/2.)*Me**2.)**(-gamma/(gamma-1.))
    
    return pr_isentropic

## @ingroup Methods-Propulsion
def pressure_ratio_shock_in_nozzle(area_ratio, gamma):
    """ Determines the lower value of pressure ratio responsible for a 
    normal shock taking place inside the nozzle
    
    Assumptions:
    yields maximium pressure ratio where shock takes place inside the nozzle, given area ratio
    
    Source:
    Unknown

    Inputs:
    area_ratio         [dimensionless]
    gamma              [dimensionless]
    
    Outputs:
    pr_shock_in_nozzle [dimensionless]    
    
    """
    
    Me = mach_area(area_ratio, gamma, False)
    M2 = normal_shock(Me, gamma)
    
    pr_shock_in_nozzle = ((area_ratio)*(((gamma+1.)/2.)**((gamma+1.)/(2.*(gamma-1))))*M2*((1.+((gamma-1.)/2.)*M2**2.)**0.5))**(-1.)
    
    return pr_shock_in_nozzle


# RCAIDE/Library/Methods/Weights/Correlation_Buildups/Common/compute_payload_weight.py
# 
# Created: Sep 2024, M. Clarke 

# ---------------------------------------------------------------------------------------------------------------------- 
#  Imports
# ----------------------------------------------------------------------------------------------------------------------
from RCAIDE.Framework.Core import Data, Units 

# ---------------------------------------------------------------------------------------------------------------------- 
# Payload
# ---------------------------------------------------------------------------------------------------------------------- 
def compute_payload_weight(vehicle, W_passenger=195 * Units.lbs, W_baggage=30 * Units.lbs):
    """ Calculate the weight of the payload and the resulting fuel mass
    
    Assumptions:
        based on FAA guidelines for weight of passengers
        
    Source: 
        N/A
        
    Inputs:
        TOW -                                                              [kilograms]
        W_empty - Operating empty weight of the aircraft                  [kilograms]
        num_pax - number of passengers on the aircraft                     [dimensionless]
        W_cargo - weight of cargo being carried on the aircraft           [kilogram]
        W_passenger - weight of each passenger on the aircraft            [kilogram]
        W_baggage - weight of the baggage for each passenger              [kilogram]
    
    Outputs:
        output - a data dictionary with fields:
            payload - weight of the passengers plus baggage and paid cargo [kilograms]
            pax - weight of all the passengers                             [kilogram]
            bag - weight of all the baggage                                [kilogram]
            fuel - weight of the fuel carried                              [kilogram]
            empty - operating empty weight of the aircraft                 [kilograms]
               
    Properties Used:
        N/A
    """

    # process
    num_pax    = vehicle.passengers
    W_pax      = W_passenger * num_pax
    W_bag      = W_baggage * num_pax
    W_payload  = W_pax + W_bag + vehicle.mass_properties.cargo

    # packup outputs
    output              = Data()
    output.total        = W_payload
    output.passengers   = W_pax
    output.baggage      = W_bag
    output.cargo        = vehicle.mass_properties.cargo

    return output
